#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from math import ceil, log
from pathlib import Path
from traceback import format_exc

from reclaimer import data_extraction
from reclaimer.h3.constants import h3_tag_class_fcc_to_ext, FORMAT_NAME_MAP
from reclaimer.h3.util import HALO3_SHARED_MAP_TYPES, get_h3_pixel_bytes_size,\
     int_to_fourcc
from reclaimer.h3.handler import Halo3Handler
from reclaimer.h3.constants import h3_tag_class_fcc_to_ext
from reclaimer.meta.gen3_resources.bitmap import s_tag_d3d_texture_def,\
     s_tag_d3d_texture_interleaved_def
from reclaimer.meta.wrappers.halo_map import HaloMap
from reclaimer.meta.wrappers.string_id_manager import StringIdManager
from reclaimer.meta.wrappers.tag_index_manager import TagIndexManager
from reclaimer.meta.wrappers.rawdata_manager import RawdataManager
from reclaimer.meta.wrappers.tag_index_converters import h3_to_h1_tag_index

from supyr_struct.util import is_path_empty

from arbytmap.format_defs import VALID_FORMATS

__all__ = ("Halo3Map", )


def get_bitmap_pixel_data(halo_map, bitm_meta, bitmap_index):
    n_assets = bitm_meta.zone_assets_normal.STEPTREE
    i_assets = bitm_meta.zone_assets_interleaved.STEPTREE
    bitmap = bitm_meta.bitmaps.STEPTREE[bitmap_index]
    rawdata_manager = halo_map.rawdata_manager

    tag_rsrc_idx = tag_rsrc_info = d3d_texture = None
    pixel_data = ()
    if bitmap.interleaved_asset_index in range(len(i_assets)):
        tag_rsrc_idx = i_assets[bitmap.interleaved_asset_index].idx
        d3d_def = s_tag_d3d_texture_interleaved_def
    else:
        tag_rsrc_idx = n_assets[bitmap_index].idx
        d3d_def = s_tag_d3d_texture_def

    if tag_rsrc_idx is not None:
        fixup_data, _, __ = rawdata_manager.get_tag_resource_fixup(tag_rsrc_idx)
        tag_rsrc_info = rawdata_manager.get_tag_resource(tag_rsrc_idx)

        if fixup_data and tag_rsrc_info:
            d3d_texture = d3d_def.build(rawdata=fixup_data)
            pixel_data = rawdata_manager.get_tag_resource_data(
                tag_rsrc_idx,
                d3d_texture.primary_page_data.size,
                d3d_texture.secondary_page_data.size)

    return d3d_texture, pixel_data


def inject_bitmap_data(halo_map, bitm_meta):
    processed_pixel_data = bitm_meta.processed_pixel_data
    bitmaps = bitm_meta.bitmaps.STEPTREE
    n_assets = bitm_meta.zone_assets_normal.STEPTREE
    i_assets = bitm_meta.zone_assets_interleaved.STEPTREE

    processed_pixel_data.data = bytearray()
    for i in range(len(bitmaps)):
        bitmap = bitmaps[i]
        bitmap.pixels_offset = len(processed_pixel_data.data)
        try:
            fmt = FORMAT_NAME_MAP[bitmap.format.data]
        except (KeyError, IndexError):
            # unknown bitmap format
            continue

        if fmt not in VALID_FORMATS:
            continue
        d3d_texture, pixel_data_chunks = get_bitmap_pixel_data(
            halo_map, bitm_meta, i)
        if not pixel_data_chunks:
            continue
        elif len(pixel_data_chunks) == 1:
            pixel_data_chunks.append(b'')

        tex_stride = get_h3_pixel_bytes_size(
            fmt, 32, 32, 1, 0, bitmap.format_flags.tiled)

        if hasattr(d3d_texture, "texture1"):
            # slice out the main image and mips for each of the
            # interleaved bitmaps(their storage is very confusing)
            pri_chunk, sec_chunk = pixel_data_chunks[: 2]
            main_chunk, mips_chunk = bytearray(), bytearray()
            tex_num = int(bool(bitmap.interleaved_index))
            tex_page0 = d3d_texture.texture0.tex_page_index
            tex_page1 = d3d_texture.texture1.tex_page_index
            if tex_num:
                tex_page, other_tex_page = tex_page1, tex_page0
            else:
                tex_page, other_tex_page = tex_page0, tex_page1

            if tex_page0 == 0 and tex_page1 == 1:
                # idky this is needed, but without it the second textures
                # mipmaps will end up getting used as the first textures
                # main image, and the first textures main image and mipmaps
                # will end up getting used as the second textures mipmaps.
                tex_num = (tex_num + 1) % 2

            tex_skip_size = tex_num * tex_stride // 2

            mips_start = 0
            if tex_page == 0:
                mips_start = bitmap.mipmap_data_off
                main_chunk += pri_chunk[tex_skip_size: tex_skip_size + mips_start]
            elif not other_tex_page:
                main_chunk += sec_chunk
            elif tex_num:
                main_chunk += sec_chunk[len(sec_chunk) // 2: ]
            else:
                main_chunk += sec_chunk[: len(sec_chunk) // 2]

            mips_chunk += pri_chunk[mips_start + tex_skip_size: ]

            pixel_data_chunks = [mips_chunk, main_chunk]

        bitmap.pixels_data_size = 0
        for pixel_data in reversed(pixel_data_chunks):
            bitmap.pixels_data_size += len(pixel_data)
            processed_pixel_data.data += pixel_data


class Halo3Map(HaloMap):
    '''Generation 3 map'''
    rawdata_manager = None

    shared_map_names = ()

    root_tags = {}

    tag_defs_module = "reclaimer.h3.defs"
    tag_classes_to_load = tuple(sorted(h3_tag_class_fcc_to_ext.keys()))

    handler_class = Halo3Handler

    string_id_set_offsets = (
        (0x4B7, 0xC11), (0x0, 0x4B7), (0x0, 0xA7D),
        (0x0, 0xB0F),   (0x0, 0xBAF), (0x0, 0xB63),
        (0x0, 0xBBF),   (0x0, 0xBF0), (0x0, 0xC04))

    data_extractors = data_extraction.h3_data_extractors

    def __init__(self, maps=None):
        self.root_tags = {}
        HaloMap.__init__(self, maps)
        self.setup_tag_headers()

    def basic_deprotection(self):
        if self.tag_index is None:
            return

        for b in self.tag_index.tag_index:
            if not b.path:
                b.path = self.map_header.map_name

        HaloMap.basic_deprotection(self)

    def get_meta_descriptor(self, tag_cls):
        tagdef = self.defs.get(tag_cls)
        if tagdef is not None:
            return tagdef.descriptor[1]

    def get_root_tag(self, tag_id_or_cls):
        if isinstance(tag_id_or_cls, str):
            for b in self.orig_tag_index.root_tags:
                if b.tag_class.enum_name == tag_id_or_cls:
                    tag_id_or_cls = b.id
                    break

        if isinstance(tag_id_or_cls, int):
            tag_id_or_cls &= 0xFFff
            if tag_id_or_cls not in self.root_tags:
                self.load_root_tags(tag_id_or_cls)

        return self.root_tags.get(tag_id_or_cls)

    def load_root_tags(self, tag_ids_to_load=()):
        tag_classes_to_load_by_ids = {}
        for b in self.orig_tag_index.root_tags:
            tag_classes_to_load_by_ids[b.id] = b.tag_class.enum_name

        if not tag_ids_to_load:
            tag_ids_to_load = set(tag_classes_to_load_by_ids).difference(
                ("scenario", "globals"))

        for tag_id in tag_ids_to_load:
            tag_cls = tag_classes_to_load_by_ids[tag_id]
            meta = self.get_meta(tag_id)
            if meta:
                self.root_tags[tag_id & 0xFFff] = meta
                if tag_cls:
                    self.root_tags[tag_cls] = meta

    def load_map(self, map_path, **kwargs):
        HaloMap.load_map(self, map_path, **kwargs)
        self.tag_index = h3_to_h1_tag_index(self.map_header, self.tag_index)
        self.basic_deprotection()

        tag_index_array = self.tag_index.tag_index
        self.tag_index_manager = TagIndexManager(tag_index_array)
        self.string_id_manager = StringIdManager(
            self.map_header.strings.string_id_table,
            self.string_id_set_offsets,
            )
        for sect in self.map_header.sections:
            self.map_pointer_converter.add_page_info(
                sect.virtual_address, sect.file_offset, sect.size)

        for part in self.map_header.partitions:
            self.map_pointer_converter.add_page_info(
                part.load_address, part.file_offset, part.size)

        self.load_root_tags()

        play_meta = self.root_tags.get("cache_file_resource_layout_table")
        if play_meta:
            self.shared_map_names = list(
                b.map_path.lower().replace('\\', '/').split("/")[-1].split(".")[0]
                for b in play_meta.external_cache_references.STEPTREE)

        self.rawdata_manager = RawdataManager(self)

        map_type = self.map_header.map_type.data
        if map_type >= 2 and map_type < 5:
            self.is_resource = True
            ext_cache_name = HALO3_SHARED_MAP_TYPES[map_type - 2]

            for map_name, halo_map in self.maps.items():
                # update each map's rawdata_manager so each one
                # knows what name the shared caches are named.
                halo_map.rawdata_manager.add_shared_map_name(
                    ext_cache_name, self.map_header.map_name)

        self.clear_map_cache()

    def get_resource_map_paths(self, maps_dir=""):
        play_meta = self.root_tags.get("cache_file_resource_layout_table")
        if not play_meta:
            return {}

        map_paths = {name.split(".")[0]: None for name in self.shared_map_names}
        if not is_path_empty(maps_dir):
            maps_dir = Path(maps_dir)
        else:
            maps_dir = self.filepath.parent

        # detect/ask for the map paths for the resource maps
        for map_name in sorted(map_paths):
            map_path = str(maps_dir.joinpath("%s.map" % map_name))
            if self.maps.get(map_name) is not None:
                map_paths[map_name] = self.maps[map_name].filepath
            elif os.path.exists(map_path):
                map_paths[map_name] = map_path

        return map_paths

    def get_meta(self, tag_id, reextract=False, **kw):
        if tag_id is None or self.map_header.map_type.data > 2:
            # shared maps don't have a tag index
            return

        # if we are given a 32bit tag id, mask it off
        tag_id &= 0xFFff
        tag_index_ref = self.tag_index_manager.get_tag_index_ref(tag_id)
        if tag_index_ref is None:
            return

        tag_cls = None
        full_tag_cls_name = tag_index_ref.class_1.enum_name
        if full_tag_cls_name not in ("<INVALID>", "NONE"):
            tag_cls = int_to_fourcc(tag_index_ref.class_1.data)

        desc = self.get_meta_descriptor(tag_cls)
        if desc is None or tag_cls is None:
            return
        elif not reextract and self.root_tags.get(full_tag_cls_name):
            return self.root_tags[full_tag_cls_name]

        block = [None]
        try:
            # read the meta data from the map
            offset = self.map_pointer_converter.v_ptr_to_f_ptr(
                tag_index_ref.meta_offset)

            desc['TYPE'].parser(
                desc, parent=block, attr_index=0,
                rawdata=self.map_data, offset=offset,
                tag_index_manager=self.tag_index_manager,
                map_string_id_manager=self.string_id_manager,
                map_pointer_converter=self.map_pointer_converter,)
        except Exception:
            print(format_exc())
            if kw.get("allow_corrupt"):
                return block[0]
            return

        try:
            self.record_map_cache_read(tag_id, 0)
            if self.map_cache_over_limit():
                self.clear_map_cache()

            self.inject_rawdata(block[0], tag_cls, tag_index_ref)
        except Exception:
            print(format_exc())
            if not kw.get("allow_corrupt"):
                meta = None

        return block[0]

    def inject_rawdata(self, meta, tag_cls, tag_index_ref):
        # get some rawdata that would be pretty annoying to do in the parser
        if tag_cls == "bitm":
            inject_bitmap_data(self, meta)

        return meta

    def meta_to_tag_data(self, meta, tag_cls, tag_index_ref, **kwargs):
        if tag_cls == "bitm":
            # make sure all bitmaps have data, and if they don't we don't
            # consider this as a valid bitmap, so it shouldnt be a tag.
            if not meta.processed_pixel_data.data:
                return None
            del meta.zone_assets_normal.STEPTREE[:]
            del meta.zone_assets_interleaved.STEPTREE[:]

        elif tag_cls in ("jmad", "mode", "pmdf", "sLdT", "sbsp", "snd!"):
            # cannot extract these until rawdata extraction is implemented
            return None

        return meta

    def generate_map_info_string(self):
        string = HaloMap.generate_map_info_string(self)

        string += """

Tag index:
    tag count           == %s
    tag types count     == %s
    root tags count     == %s
    index array pointer == %s""" % (
        self.orig_tag_index.tag_count,
        self.orig_tag_index.tag_types_count,
        self.orig_tag_index.root_tags_count,
        self.tag_index.tag_index_offset - self.map_magic)

        for arr_name, arr in (("Partitions", self.map_header.partitions),
                              ("Sections", self.map_header.sections),):
            string += "\n%s:\n" % arr_name
            names = ("debug", "resource", "tag", "locale")\
                    if arr.NAME_MAP else range(len(arr))
            for name in names:
                section = arr[name]
                string += """
    %s:
        address == %s
        size    == %s
        offset  == %s""" % (
            name, section[0], section[1], section.file_offset)
        return string



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


rmdf_method_shader_option = Struct("shader_option",
    h3_string_id("type"),
    h3_dependency("option"),
    h3_string_id("unknown_0", VISIBLE=False),
    h3_string_id("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=28
    )


rmdf_method = Struct("method",
    h3_string_id("type"),
    h3_reflexive("shader_options", rmdf_method_shader_option),
    h3_string_id("unknown_0", VISIBLE=False),
    h3_string_id("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=24
    )


rmdf_unknown_1_unknown_1_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


rmdf_unknown_1_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_1", rmdf_unknown_1_unknown_1_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


rmdf_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_1", rmdf_unknown_1_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


rmdf_unknown_2 = Struct("unknown_2",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


rmdf_body = Struct("tagdata",
    h3_dependency("unknown_0", VISIBLE=False),
    h3_reflexive("methods", rmdf_method),
    h3_reflexive("unknown_1", rmdf_unknown_1),
    h3_reflexive("unknown_2", rmdf_unknown_2),
    h3_dependency("unknown_3", VISIBLE=False),
    h3_dependency("unknown_4", VISIBLE=False),
    BytesRaw("unknown_5", SIZE=8, VISIBLE=False),
    ENDIAN=">", SIZE=92
    )


def get():
    return rmdf_def

rmdf_def = TagDef("rmdf",
    h3_blam_header('rmdf'),
    rmdf_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["rmdf"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: -DeToX-
# 	Created layout of plugin
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


dctr_body = Struct("tagdata",
    h3_dependency("model"),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    SInt32("unknown_1", VISIBLE=False),
    h3_dependency("texture"),
    SInt16("affects_visibility"),
    SInt16("unknown_2"),
    color_rgb_float("color"),
    BytesRaw("unknown_3", SIZE=20, VISIBLE=False),
    Float("brightness_base"),
    Float("brightness_shadow"),
    BytesRaw("unknown_4", SIZE=36, VISIBLE=False),
    ENDIAN=">", SIZE=128
    )


def get():
    return dctr_def

dctr_def = TagDef("dctr",
    h3_blam_header('dctr'),
    dctr_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["dctr"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DarkShallFall
# 	Easy enough...
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


wgtz_screen_widget = Struct("screen_widget",
    h3_dependency("widget"),
    ENDIAN=">", SIZE=16
    )


wgtz_body = Struct("tagdata",
    h3_dependency("shared_ui_globals"),
    h3_dependency("editable_settings"),
    h3_dependency("matchmaking_hopper_strings"),
    h3_reflexive("screen_widgets", wgtz_screen_widget),
    ENDIAN=">", SIZE=60
    )


def get():
    return wgtz_def

wgtz_def = TagDef("wgtz",
    h3_blam_header('wgtz'),
    wgtz_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["wgtz"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
THIS DEFINITION IS INCORRECT BECAUSE THE UNIT STRUCTURE IS DIFFERENT THAN HALO'S
'''
from ...hek.defs.vehi import *
from .obje import *
from .unit import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=1)

vehi_attrs = dict(vehi_attrs)
vehi_attrs[1] = SEnum16('type', *vehicle_types)

vehi_body = Struct("tagdata",
    obje_attrs,
    unit_attrs,
    vehi_attrs,
    SIZE=1008,
    )

#def get():
#    return vehi_def
del get

vehi_def = TagDef("vehi",
    blam_header_stubbs('vehi'),
    vehi_body,

    ext=".vehicle", endian=">", tag_cls=ObjeTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .shdr import *
from supyr_struct.defs.tag_def import TagDef
from .objs.shdr import ShdrTag

chicago_4_stage_maps = Struct("four_stage_map",
    Bool16("flags" ,
        "unfiltered",
        "alpha_replicate",
        "u_clamped",
        "v_clamped",
        ),

    Pad(42),
    SEnum16("color_function", *blend_functions),
    SEnum16("alpha_function", *blend_functions),

    Pad(36),
    #shader transformations
    Float("map_u_scale"),
    Float("map_v_scale"),
    Float("map_u_offset"),
    Float("map_v_offset"),
    float_deg("map_rotation"),  # degrees
    float_zero_to_one("map_bias"),  # [0,1]
    dependency("bitmap", "bitm"),

    #shader animations
    Pad(40),
    Struct("u_animation", INCLUDE=anim_src_func_per_pha_sca),
    Struct("v_animation", INCLUDE=anim_src_func_per_pha_sca),
    Struct("rotation_animation", INCLUDE=anim_src_func_per_pha_sca_rot),

    QStruct("rotation_center", INCLUDE=xy_float),
    SIZE=220,
    )

schi_attrs = Struct("schi_attrs",
    # Shader Properties
    Struct("chicago_shader",
        UInt8("numeric_counter_limit",
            MIN=0, MAX=255, SIDETIP="[0,255]"),  # [0,255]

        Bool8("chicago_shader_flags", *trans_shdr_properties),
        SEnum16("first_map_type", *trans_shdr_first_map_type),
        SEnum16("framebuffer_blend_function", *framebuffer_blend_functions),
        SEnum16("framebuffer_fade_mode", *render_fade_mode),
        SEnum16("framebuffer_fade_source", *function_outputs),
        Pad(2),
        ),

    #Lens Flare
    float_wu("lens_flare_spacing"),  # world units
    dependency("lens_flare", "lens"),
    reflexive("extra_layers", extra_layers_block, 4,
        DYN_NAME_PATH='.filepath'),
    reflexive("maps", chicago_4_stage_maps, 4,
        DYN_NAME_PATH='.bitmap.filepath'),
    Bool32("extra_flags",
        "dont_fade_active_camouflage",
        "numeric_countdown_timer"
        ),
    SIZE=68
    )

schi_body = Struct("tagdata",
    shdr_attrs,
    schi_attrs,
    SIZE=108
    )


def get():
    return schi_def

schi_def = TagDef("schi",
    blam_header('schi'),
    schi_body,

    ext=".shader_transparent_chicago",
    endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

#Written by gbMichelle with help from the adjutant and assembly plugin files
from ..common_descs import *
from .objs.tag import *
from reclaimer.hek.defs.bitm import sequence
from supyr_struct.defs.tag_def import TagDef


sequence_related = Struct("sequence thing",
    ascii_str32("sequence name"),
    UInt16("idx"),
    SIZE=40,
    )


bitmap = Struct("bitmap",
    UEnum32('bitm id', ('bitm', 'bitm'), DEFAULT='bitm', EDITABLE=False),
    UInt16("width", SIDETIP="pixels", EDITABLE=False),
    UInt16("height", SIDETIP="pixels", EDITABLE=False),
    UInt8("depth", SIDETIP="pixels", EDITABLE=False),
    Bool8("format flags",
        "delete from cache file",
        "bitmap create attempted",
        "unknown",
        "tiled",
        ),
    SEnum16("type",
        "texture 2d",
        "texture 3d",
        "cubemap",
        "multipage 2d",
        EDITABLE=False
        ),
    SEnum16("format",
        "a8",
        "y8",
        "ay8",
        "a8y8",
        "unused4",
        "unused5",
        ("r5g6b5", 6),
        "unused7",
        ("a1r5g5b5", 8),
        ("a4r4g4b4", 9),
        ("x8r8g8b8", 10),
        ("a8r8g8b8", 11),
        "unused12",
        "unused13",
        ("dxt1", 14),
        ("dxt3", 15),
        ("dxt5", 16),
        ("p8 bump", 17),
        ("p8", 18),
        ("argbfp32", 19),
        ("rgbfp32", 20),
        ("rgbfp16", 21),
        ("v8u8", 22),
        "unused23",
        "unused24",
        "unused25",  # ui\halox\main_menu.bkd.bitmap is set to this
        #              and is palettized with dimensions 231 x 1 x 1
        "unused26",
        "unused27",
        "unused28",
        "unused29",
        "unused30",
        "unused31",
        "unused32",
        ("dxn", 33),
        ("ctx1", 34),
        ("dxt3a", 35),
        ("dxt3y", 36),
        ("dxt5a", 37),
        ("dxt5y", 38),
        ("dxt5ay", 39),
        "unused40",
        "unused41",
        "unused42",
        "unused43",
        "unused44",
        ),
    Bool16("flags",
        "power of 2 dim",
        "compressed",
        "palletized",
        "swizzled",
        "linear",
        "v16u16",
        "mip map debug level",
        "prefer low detail",
        ("always on", 1<<9),
        ("interlaced", 1<<12),
        ),
    UInt16("registration point x"),
    UInt16("registration point y"),
    UInt8("mipmaps"),
    UInt8("unknown0"),
    SInt8("interleaved asset index"),
    SInt8("interleaved index"),

    SInt32("pixels offset"),     # only valid in tag form and more than 1 bitmap
    SInt32("pixels data size"),  # only valid in tag form and more than 1 bitmap
    SInt32("unknown1"),  # always 0?
    SInt32("mipmap data off"),  # size of main image if mipmaps > 0, else 0
    SInt32("unknown2"),  # always -1?
    SInt32("unknown3"),  # always 0?
    SIZE=48,
    )


bitm_body = Struct("tagdata",
    SInt16("unknown1"),
    SInt16("unknown2"),
    SInt16("unknown3"),
    SInt16("unknown4"),
    Float("blur filter size", MIN=0.0, MAX=10.0, SIDETIP="[0,10] pixels"),
    float_neg_one_to_one("alpha bias"),
    SInt16("unknown5"),
    Bool16("flags",
        "enable diffusion dithering",
        "disable height map compression",
        "uniform sprite sequences",
        "sprite bug fix",
        "use sharp bump filter",
        "unused", # Used for decal bumps
        "use clamped bump filter",
        "invert detail fade",
        "swap x y vector components",
        "convert from signed",
        "convert to signed",
        "import mipmap chains",
        "intentionally true color"
        ),
    reflexive("sequence related", sequence_related, 256, DYN_NAME_PATH='.sequence_name'), #0x14
    reflexive("unknown sequences", sequence, 256, DYN_NAME_PATH='.sequence_name'), #0x20
    rawdata_ref("compressed color plate data"),
    rawdata_ref("processed pixel data"),
    reflexive("sequences", sequence, 256, DYN_NAME_PATH='.sequence_name'),
    reflexive("bitmaps", bitmap, 2048),

    rawdata_ref("unknown6"),
    Pad(12),
    reflexive("zone assets normal", zone_asset_struct),
    reflexive("zone assets interleaved", zone_asset_struct),

    ENDIAN=">", SIZE=164, WIDGET=Halo3BitmapTagFrame,
    )


def get():
    return bitm_def

bitm_def = TagDef("bitm",
    h3_blam_header('bitm'),
    bitm_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["bitm"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


rmop_unknown = Struct("unknown",
    h3_string_id("type"),
    BytesRaw("unknown_0", SIZE=8, VISIBLE=False),
    h3_dependency("unknown_1", VISIBLE=False),
    BytesRaw("unknown_2", SIZE=44, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=72
    )


rmop_body = Struct("tagdata",
    h3_reflexive("unknown", rmop_unknown),
    ENDIAN=">", SIZE=12
    )


def get():
    return rmop_def

rmop_def = TagDef("rmop",
    h3_blam_header('rmop'),
    rmop_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["rmop"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

lsnd_track_output_effect = (
    "none",
    "output_front_speakers",
    "output_rear_speakers",
    "output_center_speakers",
    )


lsnd_track = Struct("track",
    h3_string_id("name"),
    Bool32("flags",
        "fade_in_at_start",
        "fade_out_at_stop",
        ),
    Float("gain"),
    Float("fade_in_duration"),
    Float("fade_out_duration"),
    h3_dependency("in"),
    h3_dependency("loop"),
    h3_dependency("out"),
    h3_dependency("alternate_loop"),
    h3_dependency("alternate_out"),
    SEnum16("output_effect", *lsnd_track_output_effect),
    SInt16("unknown", VISIBLE=False),
    h3_dependency("alternate_transition_in"),
    h3_dependency("alternate_transition_out"),
    Float("alternate_crossfade_duration"),
    Float("alternate_fade_out_duration"),
    ENDIAN=">", SIZE=144
    )


lsnd_detail_sound = Struct("detail_sound",
    h3_string_id("name"),
    h3_dependency("sound"),
    QStruct("random_period_bounds", INCLUDE=from_to),
    Float("unknown", VISIBLE=False),
    Bool32("flags",
        "dont_play_with_alternate",
        "dont_play_without_alternate",
        "start_immediately_with_loop",
        ),
    from_to_rad("yaw_bounds"),
    from_to_rad("pitch_bounds"),
    QStruct("distance_bounds", INCLUDE=from_to),
    ENDIAN=">", SIZE=60
    )


lsnd_body = Struct("tagdata",
    Bool32("flags",
        "deafening_to_ais",
        "not_a_loop",
        "stops_music",
        "always_spatialize",
        "synchronize_playback",
        "synchronize_tracks",
        "fake_spatialization_with_distance",
        "combine_all_3d_playback",
        ),
    Float("marty_s_music_time"),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    h3_dependency("unused"),
    SEnum16("sound_class", *snd__sound_class),
    SInt16("unknown_3", VISIBLE=False),
    h3_reflexive("tracks", lsnd_track),
    h3_reflexive("detail_sounds", lsnd_detail_sound),
    ENDIAN=">", SIZE=64
    )


def get():
    return lsnd_def

lsnd_def = TagDef("lsnd",
    h3_blam_header('lsnd'),
    lsnd_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["lsnd"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


cisc_puppet_unknown_2 = Struct("unknown_2",
    BytesRaw("unknown_0", SIZE=40, VISIBLE=False),
    h3_dependency("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=56
    )


cisc_puppet = Struct("puppet",
    ascii_str32("import_name"),
    h3_string_id("name"),
    h3_string_id("variant"),
    h3_dependency("puppet_animation"),
    h3_dependency("puppet_object"),
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    SInt32("unknown_1", VISIBLE=False),
    h3_rawdata_ref("import_script"),
    h3_reflexive("unknown_2", cisc_puppet_unknown_2),
    ENDIAN=">", SIZE=116
    )


cisc_shot_lighting = Struct("lighting",
    h3_dependency("cinematic_light"),
    SInt32("owner_puppet_index"),
    h3_string_id("marker"),
    ENDIAN=">", SIZE=24
    )


cisc_shot_unknown_3_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


cisc_shot_unknown_3 = Struct("unknown_3",
    BytesRaw("unknown_0", SIZE=32, VISIBLE=False),
    h3_reflexive("unknown_1", cisc_shot_unknown_3_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=44
    )


cisc_shot_sound = Struct("sound",
    h3_dependency("sound"),
    SInt32("frame"),
    Float("unknown_0", VISIBLE=False),
    h3_string_id("unknown_1"),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    h3_string_id("unknown_3"),
    ENDIAN=">", SIZE=36
    )


cisc_shot_background_sound = Struct("background_sound",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    h3_dependency("sound"),
    SInt32("frame"),
    ENDIAN=">", SIZE=24
    )


cisc_shot_effect = Struct("effect",
    h3_dependency("effect"),
    SInt32("frame"),
    h3_string_id("marker"),
    SInt32("owner_puppet_index"),
    ENDIAN=">", SIZE=28
    )


cisc_shot_function_value = Struct("value",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    SInt32("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    Float("unknown_3", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


cisc_shot_function = Struct("function",
    SInt32("owner_puppet_index"),
    h3_string_id("target_function_name"),
    h3_reflexive("values", cisc_shot_function_value),
    ENDIAN=">", SIZE=20
    )


cisc_shot_cortana_effect = Struct("cortana_effect",
    h3_dependency("effect"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=20
    )


cisc_shot_import_script = Struct("import_script",
    SInt32("frame"),
    h3_rawdata_ref("import_script"),
    ENDIAN=">", SIZE=24
    )


cisc_shot_frame = Struct("frame",
    QStruct("position", INCLUDE=xyz_float),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    Float("fov"),
    Bool32("flags",
        "enable_depth_of_field",
        ),
    Float("near_plane"),
    Float("far_plane"),
    Float("focal_depth"),
    Float("blur_amount"),
    ENDIAN=">", SIZE=68
    )


cisc_shot = Struct("shot",
    h3_rawdata_ref("opening_import_script"),
    SInt32("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    h3_reflexive("lighting", cisc_shot_lighting),
    h3_reflexive("unknown_3", cisc_shot_unknown_3),
    h3_reflexive("sounds", cisc_shot_sound),
    h3_reflexive("background_sounds", cisc_shot_background_sound),
    h3_reflexive("effects", cisc_shot_effect),
    h3_reflexive("functions", cisc_shot_function),
    h3_reflexive("cortana_effects", cisc_shot_cortana_effect),
    h3_reflexive("import_scripts", cisc_shot_import_script),
    h3_rawdata_ref("import_script"),
    SInt32("loaded_frame_count"),
    h3_reflexive("frames", cisc_shot_frame),
    ENDIAN=">", SIZE=164
    )


cisc_texture_camera_shot_frame = Struct("frame",
    SInt32("unknown", VISIBLE=False),
    QStruct("position", INCLUDE=xyz_float),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    Float("fov"),
    Bool32("flags",
        "enable_depth_of_field",
        ),
    Float("near_plane"),
    Float("far_plane"),
    Float("focal_depth"),
    Float("blur_amount"),
    ENDIAN=">", SIZE=72
    )


cisc_texture_camera_shot = Struct("shot",
    h3_reflexive("frames", cisc_texture_camera_shot_frame),
    ENDIAN=">", SIZE=12
    )


cisc_texture_camera = Struct("texture_camera",
    h3_string_id("name"),
    h3_string_id("unknown"),
    h3_reflexive("shots", cisc_texture_camera_shot),
    ENDIAN=">", SIZE=20
    )


cisc_body = Struct("tagdata",
    h3_string_id("name"),
    ascii_str32("anchor_name"),
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_rawdata_ref("import_script_0"),
    h3_reflexive("puppets", cisc_puppet),
    h3_reflexive("shots", cisc_shot),
    h3_reflexive("texture_cameras", cisc_texture_camera),
    h3_rawdata_ref("import_script_1"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=120
    )


def get():
    return cisc_def

cisc_def = TagDef("cisc",
    h3_blam_header('cisc'),
    cisc_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["cisc"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Copypasta H2
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


snde_body = Struct("tagdata",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    SInt16("priority"),
    SInt16("unknown_1", VISIBLE=False),
    Float("room_intensity"),
    Float("room_intensity_high_frequency"),
    Float("room_rolloff"),
    Float("decay_time"),
    Float("decay_high_frequency_ratio"),
    Float("reflections_intensity", VISIBLE=False),
    Float("reflections_delay"),
    Float("reverb_intensity"),
    Float("reverb_delay"),
    Float("diffusion"),
    Float("density"),
    Float("high_frequency_refrence"),
    BytesRaw("unknown_2", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=72
    )


def get():
    return snde_def

snde_def = TagDef("snde",
    h3_blam_header('snde'),
    snde_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["snde"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.enums import *

play_raw_page_compression_codec = (
    "deflate",
    "lzx",
    ("NONE", -1)
    )

bloc_ai_propertie_leap_jump_speed = (
    "none",
    "down",
    "step",
    "crouch",
    "stand",
    "storey",
    "tower",
    "infinite",
    )

bloc_ai_propertie_size = (
    "default",
    "tiny",
    "small",
    "medium",
    "large",
    "huge",
    "immobile",
    )

bloc_attachment_change_color = (
    "none",
    "primary",
    "secondary",
    "tertiary",
    "quaternary",
    )

bloc_lightmap_shadow_mode_size = (
    "default",
    "never",
    "always",
    "unknown",
    )

bloc_metagame_propertie_classification = (
    "infantry",
    "leader",
    "hero",
    "specialist",
    "light_vehicle",
    "heavy_vehicle",
    "giant_vehicle",
    "standard_vehicle",
    )

bloc_metagame_propertie_unit = (
    "brute",
    "grunt",
    "jackal",
    "marine",
    "bugger",
    "hunter",
    "flood_infection",
    "flood_carrier",
    "flood_combat",
    "flood_pureform",
    "sentinel",
    "elite",
    "turret",
    "mongoose",
    "warthog",
    "scorpion",
    "hornet",
    "pelican",
    "shade",
    "watchtower",
    "ghost",
    "chopper",
    "mauler",
    "wraith",
    "banshee",
    "phantom",
    "scarab",
    "guntower",
    )

bloc_multiplayer_object_propertie_object_type = (
    "ordinary",
    "weapon",
    "grenade",
    "projectile",
    "powerup",
    "equipment",
    "light_land_vehicle",
    "heavy_land_vehicle",
    "flying_vehicle",
    "teleporter_2way",
    "teleporter_sender",
    "teleporter_receiver",
    "player_spawn_location",
    "player_respawn_zone",
    "hold_spawn_objective",
    "capture_spawn_objective",
    "hold_destination_objective",
    "capture_destination_objective",
    "hill_objective",
    "infection_haven_objective",
    "territory_objective",
    "vip_boundary_objective",
    "vip_destination_objective",
    "juggernaut_destination_objective",
    )

bloc_multiplayer_object_propertie_shape = (
    "none",
    "sphere",
    "cylinder",
    "box",
    )

bloc_multiplayer_object_propertie_spawn_timer_mode = (
    "on_death",
    "on_disturbance",
    )

bloc_object_type = (
    "biped",
    "vehicle",
    "weapon",
    "equipment",
    "terminal",
    "projectile",
    "scenery",
    "machine",
    "control",
    "sound_scenery",
    "crate",
    "creature",
    "giant",
    "effect_scenery",
    )

bloc_sweetener_size = (
    "small",
    "medium",
    "large",
    )

bloc_water_density = (
    "default",
    "least",
    "some",
    "equal",
    "more",
    "more_still",
    "lots_more",
    )

bmp3_blend_method = (
    "standard",
    "unknown_0",
    "unknown_1",
    "alpha",
    "overlay",
    "unknown_2",
    "lighter_color",
    "unknown_3",
    "unknown_4",
    "unknown_5",
    "inverted_alpha",
    "unknown_6",
    "unknown_7",
    "unknown_8",
    )

cntl_contrail_system_output_kind_0 = (
    "none",
    "plus",
    "times",
    )

crea_default_team = (
    "default",
    "player",
    "human",
    "covenant",
    "flood",
    "sentinel",
    "heretic",
    "prophet",
    "guilty",
    "unused9",
    "unused10",
    "unused11",
    "unused12",
    "unused13",
    "unused14",
    "unused15",
    )

crea_motion_sensor_blip_size = (
    "medium",
    "small",
    "large",
    )

gint_constant_sound_volume = (
    "silent",
    "medium",
    "loud",
    "shout",
    "quiet",
    )

gint_grenade_type = (
    "human_fragmentation",
    "covenant_plasma",
    "brute_spike",
    "firebomb",
    )

gint_item_scale = (
    "small",
    "medium",
    "large",
    "huge",
    )

gint_seat_ai_seat_type = (
    "none",
    "passenger",
    "gunner",
    "small_cargo",
    "large_cargo",
    "driver",
    )

jpt__player_response_fade_function = (
    "linear",
    "late",
    "very_late",
    "early",
    "very_early",
    "cosine",
    "zero",
    "one",
    )

jpt__player_response_response_type = (
    "shielded",
    "unshielded",
    "all",
    )

jpt__wobble_function = (
    "one",
    "zero",
    "cosine",
    "cosine_variable_period",
    "diagonal_wave",
    "diagonal_wave_variable_period",
    "slide",
    "slide_variable_period",
    "noise",
    "jitter",
    "wander",
    "spark",
    )

phmo_node_edge_constraint_type = (
    "hinge",
    "limited_hinge",
    "ragdoll",
    "stiff_spring",
    "ball_and_socket",
    "prismatic",
    )

pmdf_meshe_index_buffer_type = (
    "point_list",
    "line_list",
    "line_strip",
    "triange_list",
    "triange_fan",
    "triange_strip",
    )

pmdf_meshe_prt_type = (
    "none",
    "ambient",
    "linear",
    "quadratic",
    )

pmdf_meshe_vertex_type = (
    "world",
    "rigid",
    "skinned",
    "particle_model",
    "flat_world",
    "flat_rigid",
    "flat_skinned",
    "screen",
    "debug",
    "transparent",
    "particle",
    "contrail",
    "light_volume",
    "chud_simple",
    "chud_fancy",
    "decorator",
    "tiny_position",
    "patchy_fog",
    "water",
    "ripple",
    "implicit",
    "beam",
    )

proj_damage_reporting_type = (
    "guardians",
    "falling_damage",
    "collision",
    "melee",
    "explosion",
    "magnum",
    "plasma_pistol",
    "needler",
    "mauler",
    "smg",
    "plasma_rifle",
    "battle_rifle",
    "carbine",
    "shotgun",
    "sniper_rifle",
    "beam_rifle",
    "assault_rifle",
    "spiker",
    "fuel_rod_cannon",
    "missile_pod",
    "rocket_launcher",
    "spartan_laser",
    "brute_shot",
    "flamethrower",
    "sentinel_gun",
    "energy_sword",
    "gravity_hammer",
    "frag_grenade",
    "plasma_grenade",
    "spike_grenade",
    "firebomb_grenade",
    "flag",
    "bomb",
    "bomb_explosion",
    "ball",
    "machinegun_turret",
    "plasma_cannon",
    "plasma_mortar",
    "plasma_turret",
    "banshee",
    "ghost",
    "mongoose",
    "scorpion",
    "scorpion_gunner",
    "spectre",
    "spectre_gunner",
    "warthog",
    "warthog_gunner",
    "warthog_gauss_turret",
    "wraith",
    "wraith_gunner",
    "tank",
    "chopper",
    "hornet",
    "mantis",
    "prowler",
    "sentinel_beam",
    "sentinel_rpg",
    "teleporter",
    "tripmine",
    "elephant_turret",
    )

scnr_light_volume_type_1 = (
    "sphere",
    "projective",
    )

sily_text_value_pair_expected_value_type = (
    "integer_index",
    "stringid_reference",
    "incremental",
    )

snd__sound_class = (
    "projectile_impact",
    "projectile_detonation",
    "projectile_flyby",
    "projectile_detonation_lod",
    "weapon_fire",
    "weapon_ready",
    "weapon_reload",
    "weapon_empty",
    "weapon_charge",
    "weapon_overheat",
    "weapon_idle",
    "weapon_melee",
    "weapon_animation",
    "object_impacts",
    "particle_impacts",
    "weapon_fire_lod",
    "unused1_impacts",
    "unused2_impacts",
    "unit_footsteps",
    "unit_dialog",
    "unit_animation",
    "unit_unused",
    "vehicle_collision",
    "vehicle_engine",
    "vehicle_animation",
    "vehicle_engine_lod",
    "device_door",
    "device_unused0",
    "device_machinery",
    "device_stationary",
    "device_unused1",
    "device_unused2",
    "music",
    "ambient_nature",
    "ambient_machinery",
    "ambient_stationary",
    "huge_ass",
    "object_looping",
    "cinematic_music",
    "unknown_unused0",
    "unknown_unused1",
    "ambient_flock",
    "no_pad",
    "no_pad_stationary",
    "mission_unused0",
    "cortana_mission",
    "cortana_gravemind_channel",
    "mission_dialog",
    "cinematic_dialog",
    "scripted_cinematic_foley",
    "game_event",
    "ui",
    "test",
    "multilingual_test",
    )

ugh__language_language = (
    "english",
    "japanese",
    "german",
    "french",
    "spanish",
    "latin_american_spanish",
    "italian",
    "korean",
    "chinese_traditional",
    "chinese_simplified",
    "portuguese",
    "polish",
    )

unknown_flags_16 = tuple("bit_%s" % i for i in range(16))

unknown_flags_32 = tuple("bit_%s" % i for i in range(32))

unknown_flags_8 = tuple("bit_%s" % i for i in range(8))

vehi_friction_point_model_state_destroyed = (
    "default",
    "minor_damage",
    "medium_damage",
    "major_damage",
    "destroyed",
    )

wscl_animation_definition_anchor = (
    "custom",
    "center",
    "top_center",
    "bottom_center",
    "left_center",
    "right_center",
    "top_left",
    "top_right",
    "bottom_right",
    "bottom_left",
    )

zone_map_type = (
    "single_player",
    "multiplayer",
    "main_menu",
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


rmt2_unknown_1 = Struct("unknown_1",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


rmt2_unknown_2 = Struct("unknown_2",
    BytesRaw("unknown", SIZE=28, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=28
    )


rmt2_unknown_3 = Struct("unknown_3",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


rmt2_argument = Struct("argument",
    h3_string_id("name"),
    ENDIAN=">", SIZE=4
    )


rmt2_unknown_4 = Struct("unknown_4",
    h3_string_id("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


rmt2_unknown_5 = Struct("unknown_5",
    h3_string_id("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


rmt2_shader_map = Struct("shader_map",
    h3_string_id("name"),
    ENDIAN=">", SIZE=4
    )


rmt2_body = Struct("tagdata",
    h3_dependency("vertex_shader"),
    h3_dependency("pixel_shader"),
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_1", rmt2_unknown_1),
    h3_reflexive("unknown_2", rmt2_unknown_2),
    h3_reflexive("unknown_3", rmt2_unknown_3),
    h3_reflexive("arguments", rmt2_argument),
    h3_reflexive("unknown_4", rmt2_unknown_4),
    h3_reflexive("unknown_5", rmt2_unknown_5),
    h3_reflexive("shader_maps", rmt2_shader_map),
    BytesRaw("unknown_6", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=132
    )


def get():
    return rmt2_def

rmt2_def = TagDef("rmt2",
    h3_blam_header('rmt2'),
    rmt2_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["rmt2"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .avti import *
from ...hek.defs.objs.tag import HekTag

instigators_comment = """
Controls which attacking unit type can trigger this transform.
Transforms with instigators will only be used when transform on damage is enabled.
"""

animations_comment = """
Played when the transform criteria have been met.
The actor is made braindead during the animation, and if the invincibility
flag is set, they cannot be killed once the animation begins.
"""

attachments_comment = """
The listed attachments will be added to the unit when it's transformation
begins. When specifying a single destination marker the destination marker
name will be used as-is(ex: 'marker'), however when the marker count is
more than 1, an index will be appended('marker0', 'marker1', etc.)
"""

instigator = Struct("instigator",
    dependency_os("unit", valid_units),
    SEnum16("damage_type",
        'both',
        'weapon_damage_only',
        'melee_damage_only',
        ),
    SIZE=32, COMMENT=instigators_comment
    )

attachment = Struct("attachment",
    dependency_os("object", valid_objects),
    ascii_str32("object_marker"),
    ascii_str32("destination_marker"),
    SInt16("destination_marker_count"),
    Pad(2),
    SEnum16("team_handling",
        'inherit_from_old_unit',
        'inherit_from_attacker',
        'override',
        ),
    SEnum16("team_override", *unit_teams),
    QStruct("attachment_scale", INCLUDE=from_to),
    SIZE=120
    )

avto_body = Struct("tagdata",
    Bool16("flags",
        'invincible_during_transform_out',
        ),
    Pad(2),

    Struct("transform_criteria",
        Bool16("flags",
            'transform_on_damage',
            'ignore_friendly_fire',
            'transform_on_actor_action',
            'transform_on_actor_state',
            'transform_on_shield_range',
            'transform_on_health_range',
            ),
        Pad(2),
        Bool16("actor_action",
            'none',
            'sleep',
            'alert',
            'fight',
            'flee',
            'uncover',
            'guard',
            'search',
            'wait',
            'vehicle',
            'charge',
            'obey',
            'converse',
            'avoid',
            ),
        Bool16("actor_state",
            'none',
            'sleeping',
            'alert',
            'moving_repeat_same_position',
            'moving_loop',
            'moving_loop_back_and_forth',
            'moving_loop_randomly',
            'moving_randomly',
            'guarding',
            'guarding_at_guard_position',
            'searching',
            'fleeing',
            ),
        from_to_zero_to_one("shield_range"),
        from_to_zero_to_one("health_range"),
        Pad(24),
        ),

    reflexive("transform_instigators", instigator, 16,
        DYN_NAME_PATH='.unit.filepath'),

    Struct("animation",
        ascii_str32("transform_out_anim"),
        reflexive("keyframe_actions", keyframe_action, 9,
            DYN_NAME_PATH='.effect_marker'),
        COMMENT=animation_comment
        ),

    Struct("attachments",
        Bool16("flags",
            "destroy_attachments_on_death",
            ),
        Pad(2),
        reflexive("attachments", attachment, 16,
            DYN_NAME_PATH='.object.filepath'),
        Pad(24),
        COMMENT=attachments_comment,
        ),
    SIZE=148
    )

def get():
    return avto_def

avto_def = TagDef("avto",
    blam_header_os('avto', 1),
    avto_body,

    ext=".actor_variant_transform_out", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: -DeToX-
# 	Mapped some value names.
# revision: 3		author: DeadCanadian
# 	Mapped some value names.
# revision: 4		author: Lord Zedd
# 	Updated and standardized.
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

char_general_propertie_actor_type = (
    "elite",
    "jackal",
    "grunt",
    "hunter",
    "engineer",
    "assassin",
    "player",
    "marine",
    "crew",
    "combat_form",
    "infection_form",
    "carrier_form",
    "monitor",
    "sentinel",
    "none",
    "mounted_weapon",
    "brute",
    "prophet",
    "bugger",
    "juggernaut",
    "pure_form_stealth",
    "pure_form_tank",
    "pure_form_ranged",
    "scarab",
    "guardian",
    )

char_grenades_propertie_grenade_type = (
    "frag",
    "plasma",
    "claymore",
    "firebomb",
    )

char_grenades_propertie_trajectory_type = (
    "toss",
    "lob",
    "bounce",
    )

char_movement_propertie_obstace_leap_minimum_size = (
    "none",
    "tiny",
    "small",
    "medium",
    "large",
    "huge",
    "immobile",
    )

char_weapons_propertie_special_fire_mode = (
    "none",
    "overcharge",
    "secondary_trigger",
    )

char_weapons_propertie_special_fire_situation = (
    "never",
    "enemy_visible",
    "enemy_out_of_sight",
    "strafing",
    )


char_variant_dialogue_variation = Struct("dialogue_variation",
    h3_dependency("dialogue"),
    h3_string_id("name"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=24
    )


char_variant = Struct("variant",
    h3_string_id("variant_name"),
    SInt16("variant_index"),
    SInt16("unknown", VISIBLE=False),
    h3_reflexive("dialogue_variations", char_variant_dialogue_variation),
    ENDIAN=">", SIZE=20
    )


char_unit_dialogue_dialogue_variation = Struct("dialogue_variation",
    h3_dependency("dialogue"),
    h3_string_id("name"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=24
    )


char_unit_dialogue = Struct("unit_dialogue",
    h3_reflexive("dialogue_variations", char_unit_dialogue_dialogue_variation),
    ENDIAN=">", SIZE=12
    )


char_general_propertie = Struct("general_propertie",
    Bool32("flags",
        "swarm",
        "flying",
        "dual_wields",
        "uses_gravemind",
        ("cannot_swap_weapons_with_player", 1 << 5),
        "does_not_initially_backpack",
        "revive_on_death",
        ),
    SEnum16("actor_type", *char_general_propertie_actor_type),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("unknown_2", VISIBLE=False),
    BytesRaw("unknown_3", SIZE=8, VISIBLE=False),
    Float("scariness"),
    SInt16("unknown_4", VISIBLE=False),
    SInt16("unknown_5", VISIBLE=False),
    ENDIAN=">", SIZE=28
    )


char_vitality_propertie = Struct("vitality_propertie",
    Bool32("vitality_flags",
        "can_revive_on_death",
        ),
    Float("normal_body_vitality"),
    Float("normal_shield_vitality"),
    Float("legendary_body_vitality"),
    Float("legendary_shield_vitality"),
    Float("body_recharge_fraction"),
    Float("soft_ping_shield_threshold"),
    Float("soft_ping_noshield_threshold"),
    Float("soft_ping_minimum_interrupt_time"),
    Float("hard_ping_shield_threshold"),
    Float("hard_ping_noshield_threshold"),
    Float("hard_ping_minimum_interrupt_time"),
    Float("current_damage_decay_delay"),
    Float("current_damage_decay_time"),
    Float("recent_damage_decay_delay"),
    Float("recent_damage_decay_time"),
    Float("body_recharge_delay_time"),
    Float("body_recharge_time"),
    Float("shield_recharge_delay_time"),
    Float("shield_recharge_time"),
    Float("stun_threshold"),
    QStruct("stun_time_bounds", INCLUDE=from_to),
    Float("extended_shield_damage_threshold"),
    Float("extended_body_damage_threshold"),
    Float("suicide_radius"),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    h3_dependency("backup_weapon"),
    ENDIAN=">", SIZE=128
    )


char_placement_propertie = Struct("placement_propertie",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    Float("few_upgrade_chance_easy"),
    Float("few_upgrade_chance_normal"),
    Float("few_upgrade_chance_heroic"),
    Float("few_upgrade_chance_legendary"),
    Float("normal_upgrade_chance_easy"),
    Float("normal_upgrade_chance_normal"),
    Float("normal_upgrade_chance_heroic"),
    Float("normal_upgrade_chance_legendary"),
    Float("many_upgrade_chance_easy"),
    Float("many_upgrade_chance_normal"),
    Float("many_upgrade_chance_heroic"),
    Float("many_upgrade_chance_legendary"),
    ENDIAN=">", SIZE=52
    )


char_perception_propertie = Struct("perception_propertie",
    SInt32("perception_flags"),
    Float("max_vision_distance"),
    float_rad("central_vision_angle"),
    float_rad("max_vision_angle"),
    float_rad("peripheral_vision_angle"),
    Float("peripheral_distance"),
    Float("hearing_distance"),
    Float("notice_projectile_chance"),
    Float("notice_vehicle_chance"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    Float("first_acknowledge_surprise_distance"),
    ENDIAN=">", SIZE=44
    )


char_look_propertie = Struct("look_propertie",
    yp_float_rad("maximum_aiming_deviation"),
    yp_float_rad("maximum_looking_deviation"),
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    float_rad("noncombat_look_delta_l"),
    float_rad("noncombat_look_delta_r"),
    float_rad("combat_look_delta_l"),
    float_rad("combat_look_delta_r"),
    QStruct("noncombat_idle_looking", INCLUDE=from_to),
    QStruct("noncombat_idle_aiming", INCLUDE=from_to),
    QStruct("combat_idle_looking", INCLUDE=from_to),
    QStruct("combat_idle_aiming", INCLUDE=from_to),
    ENDIAN=">", SIZE=80
    )


char_movement_propertie = Struct("movement_propertie",
    Bool32("movement_flags",
        ("prefer_to_combat_near_friends", 1 << 2),
        ),
    Float("pathfinding_radius"),
    Float("destination_radius"),
    Float("dive_grenade_chance"),
    SEnum16("obstace_leap_minimum_size", *char_movement_propertie_obstace_leap_minimum_size),
    SEnum16("obstace_leap_maximum_size", *char_movement_propertie_obstace_leap_minimum_size),
    SEnum16("obstace_ignore_size", *char_movement_propertie_obstace_leap_minimum_size),
    SEnum16("obstace_smashable_size", *char_movement_propertie_obstace_leap_minimum_size),
    SEnum32("jump_height", *bloc_ai_propertie_leap_jump_speed),
    Bool32("movement_hint_flags", *unknown_flags_32),
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=44
    )


char_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=24, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


char_swarm_propertie = Struct("swarm_propertie",
    SInt16("scatter_killed_count"),
    SInt16("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    Float("scatter_radius"),
    Float("scatter_distance"),
    Float("hound_min_distance"),
    Float("hound_max_distance"),
    Float("perlin_offset_scale"),
    QStruct("offset_period", INCLUDE=from_to),
    Float("perlin_idle_movement_threshold"),
    Float("perlin_combat_movement_threshold"),
    BytesRaw("unknown_2", SIZE=8, VISIBLE=False),
    ENDIAN=">", SIZE=56
    )


char_ready_propertie = Struct("ready_propertie",
    QStruct("ready_time_bounds", INCLUDE=from_to),
    ENDIAN=">", SIZE=8
    )


char_engage_propertie = Struct("engage_propertie",
    Bool32("engage_flags", *unknown_flags_32),
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    Float("crouch_danger_threshold"),
    Float("stand_danger_threshold"),
    Float("fight_danger_move_threshold"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    h3_dependency("unknown_2"),
    ENDIAN=">", SIZE=40
    )


char_charge_propertie_unknown_2 = Struct("unknown_2",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=6
    )


char_charge_propertie = Struct("charge_propertie",
    Bool32("charge_flags",
        "offhand_melee_allowed",
        ),
    Float("melee_consider_range"),
    Float("melee_chance"),
    Float("melee_attack_range"),
    Float("melee_abort_range"),
    Float("melee_attack_timeout"),
    Float("melee_attack_delay_timer"),
    QStruct("melee_leap_range", INCLUDE=from_to),
    Float("melee_leap_chance"),
    Float("ideal_leap_velocity"),
    Float("max_leap_velocity"),
    Float("melee_leap_ballistic"),
    Float("melee_delay_timer"),
    BytesRaw("unknown_0", SIZE=20, VISIBLE=False),
    h3_dependency("berserk_weapon"),
    BytesRaw("unknown_1", SIZE=20, VISIBLE=False),
    h3_reflexive("unknown_2", char_charge_propertie_unknown_2),
    ENDIAN=">", SIZE=124
    )


char_evasion_propertie = Struct("evasion_propertie",
    Float("evasion_danger_threshold"),
    Float("evasion_delay_timer"),
    Float("evasion_chance"),
    Float("evasion_proximity_threshold"),
    Float("dive_retreat_chance"),
    ENDIAN=">", SIZE=20
    )


char_cover_propertie = Struct("cover_propertie",
    Bool32("cover_flags", *unknown_flags_32),
    QStruct("hide_behind_cover_time", INCLUDE=from_to),
    Float("cover_vitality_threshold"),
    Float("cover_shield_fraction"),
    Float("cover_check_delay"),
    Float("cover_danger_threshold"),
    Float("danger_upper_threshold"),
    QStruct("cover_chance", INCLUDE=from_to),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    Float("proximity_self_preserve"),
    Float("disallow_cover_distance"),
    Float("proximity_melee_distance"),
    Float("unreachable_enemy_danger_threashold"),
    Float("scary_target_threshold"),
    BytesRaw("unknown_1", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=84
    )


char_retreat_propertie = Struct("retreat_propertie",
    Bool32("retreat_flags",
        "zig_zag_when_fleeing",
        ),
    Float("shield_threshold"),
    Float("scary_target_threshold"),
    Float("danger_threshold"),
    Float("proximity_threshold"),
    QStruct("forced_cower_time_bounds", INCLUDE=from_to),
    QStruct("cower_time_bounds", INCLUDE=from_to),
    Float("proximity_ambush_threshold"),
    BytesRaw("unknown", SIZE=20, VISIBLE=False),
    float_rad("zig_zag_angle"),
    Float("zig_zag_period"),
    Float("retreat_grenade_chance"),
    h3_dependency("backup_weapon"),
    ENDIAN=">", SIZE=88
    )


char_search_propertie = Struct("search_propertie",
    Bool32("search_flags", *unknown_flags_32),
    QStruct("search_time", INCLUDE=from_to),
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    QStruct("uncover_distance_bounds", INCLUDE=from_to),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


char_pre_search_propertie = Struct("pre_search_propertie",
    Bool32("pre_search_flags", *unknown_flags_32),
    QStruct("minimum_presearch_time", INCLUDE=from_to),
    QStruct("maximum_presearch_time", INCLUDE=from_to),
    Float("minimum_certainty_radius"),
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    QStruct("minimum_suppressing_time", INCLUDE=from_to),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=40
    )


char_idle_propertie = Struct("idle_propertie",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    QStruct("idle_pose_delay_time", INCLUDE=from_to),
    ENDIAN=">", SIZE=12
    )


char_vocalization_propertie = Struct("vocalization_propertie",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    Float("look_comment_time"),
    Float("look_long_comment_time"),
    ENDIAN=">", SIZE=12
    )


char_boarding_propertie = Struct("boarding_propertie",
    Bool32("flags",
        "airborne_boarding",
        ),
    Float("max_distance"),
    Float("abort_distance"),
    Float("max_speed"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=20
    )


char_unknown_2 = Struct("unknown_2",
    BytesRaw("unknown", SIZE=8, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=8
    )


char_unknown_3 = Struct("unknown_3",
    BytesRaw("unknown", SIZE=24, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


char_weapons_propertie_firing_pattern = Struct("firing_pattern",
    Float("rate_of_fire"),
    Float("target_tracking"),
    Float("target_leading"),
    Float("burst_origin_radius"),
    float_rad("burst_origin_angle"),
    QStruct("burst_return_length", INCLUDE=from_to),
    float_rad("burst_return_angle"),
    QStruct("burst_duration", INCLUDE=from_to),
    QStruct("burst_separation", INCLUDE=from_to),
    Float("weapon_damage_modifier"),
    float_rad("projectile_error"),
    float_rad("burst_angular_velocity"),
    float_rad("maximum_error_angle"),
    ENDIAN=">", SIZE=64
    )


char_weapons_propertie = Struct("weapons_propertie",
    Bool32("weapon_flags",
        ("must_crouch_to_shoot", 1 << 1),
        ),
    h3_dependency("weapon"),
    Float("maximum_firing_range"),
    Float("minimum_firing_range"),
    QStruct("normal_combat_range", INCLUDE=from_to),
    Float("bombardment_range"),
    Float("max_special_target_distance"),
    QStruct("timid_combat_range", INCLUDE=from_to),
    QStruct("aggressive_combat_range", INCLUDE=from_to),
    Float("super_ballistic_range"),
    QStruct("ballistic_firing_bounds", INCLUDE=from_to),
    QStruct("ballistic_fraction_bounds", INCLUDE=from_to),
    QStruct("first_burst_delay_time", INCLUDE=from_to),
    Float("surprise_delay_time"),
    Float("surprise_fire_wildly_time"),
    Float("death_fire_wildly_chance"),
    Float("death_fire_wildly_time"),
    QStruct("custom_stand_gun_offset", INCLUDE=ijk_float),
    QStruct("custom_crouch_gun_offset", INCLUDE=ijk_float),
    SEnum16("special_fire_mode", *char_weapons_propertie_special_fire_mode),
    SEnum16("special_fire_situation", *char_weapons_propertie_special_fire_situation),
    Float("special_fire_chance"),
    Float("special_fire_delay"),
    Float("special_damage_modifier"),
    float_rad("special_projectile_error"),
    QStruct("drop_weapon_loaded", INCLUDE=from_to),
    QStruct("drop_weapon_ammo", INCLUDE=from_to_sint16),
    QStruct("normal_accuracy_bounds", INCLUDE=from_to),
    Float("normal_accuracy_time"),
    QStruct("heroic_accuracy_bounds", INCLUDE=from_to),
    Float("heroic_accuracy_time"),
    QStruct("legendary_accuracy_bounds", INCLUDE=from_to),
    Float("legendary_accuracy_time"),
    h3_reflexive("firing_patterns", char_weapons_propertie_firing_pattern),
    h3_dependency("weapon_melee_damage"),
    ENDIAN=">", SIZE=224
    )


char_firing_pattern_propertie_firing_pattern = Struct("firing_pattern",
    Float("rate_of_fire"),
    Float("target_tracking"),
    Float("target_leading"),
    Float("burst_origin_radius"),
    float_rad("burst_origin_angle"),
    QStruct("burst_return_length", INCLUDE=from_to),
    float_rad("burst_return_angle"),
    QStruct("burst_duration", INCLUDE=from_to),
    QStruct("burst_separation", INCLUDE=from_to),
    Float("weapon_damage_modifier"),
    float_rad("projectile_error"),
    float_rad("burst_angular_velocity"),
    float_rad("maximum_error_angle"),
    ENDIAN=">", SIZE=64
    )


char_firing_pattern_propertie = Struct("firing_pattern_propertie",
    h3_dependency("weapon"),
    h3_reflexive("firing_patterns", char_firing_pattern_propertie_firing_pattern),
    ENDIAN=">", SIZE=28
    )


char_grenades_propertie = Struct("grenades_propertie",
    SInt32("grenades_flags"),
    SEnum16("grenade_type", *char_grenades_propertie_grenade_type),
    SEnum16("trajectory_type", *char_grenades_propertie_trajectory_type),
    SInt32("minimum_enemy_count"),
    Float("enemy_radius"),
    Float("grenade_ideal_velocity"),
    Float("grenade_velocity"),
    QStruct("grenade_range", INCLUDE=from_to),
    Float("collateral_damage_radius"),
    Float("grenade_chance"),
    Float("grenade_throw_delay"),
    Float("grenade_uncover_chance"),
    Float("anti_vehicle_grenade_chance"),
    QStruct("dropped_grenade_count", INCLUDE=from_to_sint16),
    Float("dont_drop_grenades_chance"),
    ENDIAN=">", SIZE=60
    )


char_vehicle_propertie = Struct("vehicle_propertie",
    h3_dependency("unit"),
    h3_dependency("style"),
    Bool32("vehicle_flags",
        "passengers_adopt_original_squad",
        ),
    Float("ai_pathfinding_radius"),
    Float("ai_destination_radius"),
    Float("ai_deceleration_distance"),
    Float("ai_turning_radius"),
    Float("ai_inner_turning_radius"),
    Float("ai_ideal_turning_radius"),
    float_rad("ai_banshee_steering_maximum"),
    Float("ai_max_steering_angle"),
    Float("ai_max_steering_delta"),
    Float("ai_oversteering_scale"),
    from_to_rad("ai_oversteering_bounds"),
    Float("ai_side_slip_distance"),
    Float("ai_avoidance_distance"),
    Float("ai_minimum_urgency"),
    float_rad("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    Float("ai_throttle_maximum"),
    Float("ai_goal_minimum_throttle_scale"),
    Float("ai_turn_minimum_throttle_scale"),
    Float("ai_direction_minimum_throttle_scale"),
    Float("ai_acceleration_scale"),
    Float("ai_throttle_blend"),
    Float("theoretical_max_speed"),
    Float("error_scale"),
    float_rad("ai_allowable_aim_deviation_angle"),
    Float("ai_charge_tight_angle_distance"),
    Float("ai_charge_tight_angle"),
    Float("ai_charge_repeat_timeout"),
    Float("ai_charge_look_ahead_time"),
    Float("ai_consider_distance"),
    Float("ai_charge_abort_distance"),
    Float("vehicle_ram_timeout"),
    Float("ram_paralysis_time"),
    Float("ai_cover_damage_threshold"),
    Float("ai_cover_minimum_distance"),
    Float("ai_cover_time"),
    Float("ai_cover_minimum_boost_distance"),
    Float("turtling_recent_damage_threshold"),
    Float("turtling_minimum_time"),
    Float("turtling_timeout"),
    SEnum16("obstacle_ignore_size", *char_movement_propertie_obstace_leap_minimum_size),
    SInt16("unknown_2", VISIBLE=False),
    SInt16("unknown_3", VISIBLE=False),
    SInt16("unknown_4", VISIBLE=False),
    ENDIAN=">", SIZE=208
    )


char_morph_propertie = Struct("morph_propertie",
    h3_dependency("morph_character_1"),
    h3_dependency("morph_character_2"),
    h3_dependency("morph_character_3"),
    h3_dependency("morph_muffin"),
    h3_dependency("morph_weapon_1"),
    h3_dependency("morph_weapon_2"),
    h3_dependency("morph_weapon_3"),
    BytesRaw("unknown_0", SIZE=32, VISIBLE=False),
    h3_dependency("character"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    h3_string_id("unknown_2", VISIBLE=False),
    BytesRaw("unknown_3", SIZE=60, VISIBLE=False),
    ENDIAN=">", SIZE=228
    )


char_equipment_propertie_usage_condition = Struct("usage_condition",
    SInt16("unknown_0"),
    SInt16("unknown_1"),
    Float("unknown_2"),
    Float("unknown_3"),
    ENDIAN=">", SIZE=12
    )


char_equipment_propertie = Struct("equipment_propertie",
    h3_dependency("equipment"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    Float("usage_chance"),
    h3_reflexive("usage_conditions", char_equipment_propertie_usage_condition),
    ENDIAN=">", SIZE=36
    )


char_metagame_propertie = Struct("metagame_propertie",
    Bool8("flags",
        "must_have_active_seats",
        ),
    SEnum8("unit", *bloc_metagame_propertie_unit),
    SEnum8("classification", *bloc_metagame_propertie_classification),
    SInt8("unknown_0", VISIBLE=False),
    SInt16("points"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=8
    )


char_act_attachment = Struct("act_attachment",
    h3_string_id("name"),
    h3_dependency("child_object"),
    h3_string_id("child_marker"),
    h3_string_id("parent_marker"),
    ENDIAN=">", SIZE=28
    )


char_body = Struct("tagdata",
    Bool32("character_flags", *unknown_flags_32),
    h3_dependency("parent_character"),
    h3_dependency("unit"),
    h3_dependency("creature"),
    h3_dependency("style"),
    h3_dependency("major_character"),
    h3_reflexive("variants", char_variant),
    h3_reflexive("unit_dialogue", char_unit_dialogue),
    h3_reflexive("general_properties", char_general_propertie),
    h3_reflexive("vitality_properties", char_vitality_propertie),
    h3_reflexive("placement_properties", char_placement_propertie),
    h3_reflexive("perception_properties", char_perception_propertie),
    h3_reflexive("look_properties", char_look_propertie),
    h3_reflexive("movement_properties", char_movement_propertie),
    h3_reflexive("unknown_0", char_unknown_0),
    h3_reflexive("swarm_properties", char_swarm_propertie),
    h3_reflexive("ready_properties", char_ready_propertie),
    h3_reflexive("engage_properties", char_engage_propertie),
    h3_reflexive("charge_properties", char_charge_propertie),
    h3_reflexive("evasion_properties", char_evasion_propertie),
    h3_reflexive("cover_properties", char_cover_propertie),
    h3_reflexive("retreat_properties", char_retreat_propertie),
    h3_reflexive("search_properties", char_search_propertie),
    h3_reflexive("pre_search_properties", char_pre_search_propertie),
    h3_reflexive("idle_properties", char_idle_propertie),
    h3_reflexive("vocalization_properties", char_vocalization_propertie),
    h3_reflexive("boarding_properties", char_boarding_propertie),
    BytesRaw("unknown_1", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_2", char_unknown_2),
    h3_reflexive("unknown_3", char_unknown_3),
    h3_reflexive("weapons_properties", char_weapons_propertie),
    h3_reflexive("firing_pattern_properties", char_firing_pattern_propertie),
    h3_reflexive("grenades_properties", char_grenades_propertie),
    h3_reflexive("vehicle_properties", char_vehicle_propertie),
    h3_reflexive("morph_properties", char_morph_propertie),
    h3_reflexive("equipment_properties", char_equipment_propertie),
    h3_reflexive("metagame_properties", char_metagame_propertie),
    h3_reflexive("act_attachments", char_act_attachment),
    ENDIAN=">", SIZE=468
    )


def get():
    return char_def

char_def = TagDef("char",
    h3_blam_header('char'),
    char_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["char"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .obje import *
from .unit import *
from .objs.obje import ObjeTag

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=1)

vehi_attrs = Struct("vehi_attrs",
    Bool32("flags",
        "speed_wakes_physics",
        "turn_wakes_physics",
        "driver_power_wakes_physics",
        "gunner_power_wakes_physics",
        "control_opposite_sets_brake",
        "slide_wakes_physics",
        "kills_riders_at_terminal_velocity",
        "causes_collision_damage",
        "ai_weapon_cannot_rotate",
        "ai_does_not_require_driver",
        "ai_unused",
        "ai_driver_enable",
        "ai_driver_flying",
        "ai_driver_can_sidestep",
        "ai_driver_hovering",
        ),
    SEnum16('type', *vehicle_types),

    Pad(2),
    float_wu_sec("maximum_forward_speed"),
    float_wu_sec("maximum_reverse_speed"),
    float_wu_sec_sq("speed_acceleration", UNIT_SCALE=per_sec_unit_scale),
    float_wu_sec_sq("speed_deceleration", UNIT_SCALE=per_sec_unit_scale),
    Float("maximum_left_turn"),
    Float("maximum_right_turn", SIDETIP="(should be negative)"),
    float_wu("wheel_circumference"),  # world units
    Float("turn_rate",  UNIT_SCALE=per_sec_unit_scale),
    Float("blur_speed", UNIT_SCALE=per_sec_unit_scale),
    SEnum16('A_in', *vehicle_inputs),
    SEnum16('B_in', *vehicle_inputs),
    SEnum16('C_in', *vehicle_inputs),
    SEnum16('D_in', *vehicle_inputs),

    Pad(12),
    Float("maximum_left_slide"),
    Float("maximum_right_slide"),
    Float("slide_acceleration", UNIT_SCALE=per_sec_unit_scale),
    Float("slide_deceleration", UNIT_SCALE=per_sec_unit_scale),
    Float("minimum_flipping_angular_velocity", UNIT_SCALE=per_sec_unit_scale),
    Float("maximum_flipping_angular_velocity", UNIT_SCALE=per_sec_unit_scale),

    Pad(24),
    float_deg("fixed_gun_yaw"),  # degrees
    float_deg("fixed_gun_pitch"),  # degrees

    Pad(24),
    Struct("ai",
        Float("sidestep_distance"),
        Float("destination_radius"),
        Float("avoidance_distance"),
        Float("pathfinding_radius"),
        float_sec("charge_repeat_timeout"),
        Float("strafing_abort_range"),
        from_to_rad("oversteering_bounds"),  # radians
        float_rad("steering_maximum"),  # radians
        Float("throttle_maximum"),
        float_sec("move_position_time"),
        ),

    Pad(4),
    dependency('suspension_sound', "snd!"),
    dependency('crash_sound', "snd!"),
    dependency('material_effect', "foot"),
    dependency('effect', "effe"),

    SIZE=256
    )

vehi_body = Struct("tagdata",
    obje_attrs,
    unit_attrs,
    vehi_attrs,
    SIZE=1008,
    )


def get():
    return vehi_def

vehi_def = TagDef("vehi",
    blam_header('vehi'),
    vehi_body,

    ext=".vehicle", endian=">", tag_cls=ObjeTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


nclt_unknown_0 = Struct("unknown_0",
    SInt32("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=28, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


nclt_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown_0", SIZE=16, VISIBLE=False),
    h3_dependency("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


nclt_body = Struct("tagdata",
    h3_reflexive("unknown_0", nclt_unknown_0),
    h3_reflexive("unknown_1", nclt_unknown_1),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=28
    )


def get():
    return nclt_def

nclt_def = TagDef("nclt",
    h3_blam_header('nclt'),
    nclt_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["nclt"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.scex import *
from .schi import *
from supyr_struct.defs.tag_def import TagDef

scex_body = Struct("tagdata",
    shdr_attrs,
    scex_attrs,
    SIZE=120
    )

def get():
    return scex_def

scex_def = TagDef("scex",
    blam_header_stubbs('scex'),
    scex_body,

    ext=".shader_transparent_chicago_extended", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import sin, cos

from reclaimer.hek.defs.objs.tag import HekTag

class Ant_Tag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)

        tagdata = self.data.tagdata
        tagdata.length = 0
        for vertex in tagdata.vertices.STEPTREE:
            sin_y = sin(vertex.angles.y)
            sin_p = sin(vertex.angles.p)
            cos_y = cos(vertex.angles.y)
            cos_p = cos(vertex.angles.p)

            vertex.offset.x = vertex.length * sin_p * cos_y
            vertex.offset.y = vertex.length * sin_y * sin_p
            vertex.offset.z = vertex.length * cos_p

            tagdata.length += vertex.length



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Standardized model data and updated
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


sbsp_structure_seam_unknown = Struct("unknown",
    SInt32("unknown"),
    ENDIAN=">", SIZE=4
    )


sbsp_structure_seam_seam_cluster = Struct("seam_cluster",
    SInt32("cluster_index"),
    QStruct("centroid", INCLUDE=xyz_float),
    ENDIAN=">", SIZE=16
    )


sbsp_structure_seam = Struct("structure_seam",
    Array("unknown_array", SUB_STRUCT=UInt32("unknown"), SIZE=4, VISIBLE=False),
    h3_reflexive("unknown", sbsp_structure_seam_unknown),
    h3_reflexive("seam_clusters", sbsp_structure_seam_seam_cluster),
    ENDIAN=">", SIZE=40
    )


sbsp_unknown_raw_7th = Struct("unknown_raw_7th",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_collision_material = Struct("collision_material",
    h3_dependency("shader"),
    SInt16("global_material_index"),
    SInt16("conveyor_surface_index"),
    SInt16("seam_index"),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=24
    )


sbsp_unknown_raw_3rd = Struct("unknown_raw_3rd",
    SInt8("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=1
    )


sbsp_unknown_raw_6th = Struct("unknown_raw_6th",
    SInt16("unknown_1_start_index", VISIBLE=False),
    SInt16("unknown_1_entry_count", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_unknown_1 = Struct("unknown_1",
    UInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_cluster_portal_vertice = Struct("vertice",
    QStruct("position", INCLUDE=xyz_float),
    ENDIAN=">", SIZE=12
    )


sbsp_cluster_portal = Struct("portal",
    SInt16("portal_index"),
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_2 = Struct("unknown_2",
    BytesRaw("unknown", SIZE=120, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=120
    )


sbsp_fog = Struct("fog",
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=8
    )


sbsp_camera_effect = Struct("camera_effect",
    h3_string_id("name"),
    h3_dependency("effect"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    Float("unknown_3", VISIBLE=False),
    BytesRaw("unknown_4", SIZE=8, VISIBLE=False),
    ENDIAN=">", SIZE=48
    )


sbsp_detail_object_unknown_1_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


sbsp_detail_object_unknown_1 = Struct("unknown_1",
    h3_reflexive("unknown_0", sbsp_detail_object_unknown_1_unknown_0),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


sbsp_detail_object = Struct("detail_object",
    BytesRaw("unknown_0", SIZE=28, VISIBLE=False),
    h3_reflexive("unknown_1", sbsp_detail_object_unknown_1),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=52
    )


sbsp_cluster_collision_mopp_code_data = Struct("data",
    UInt8("data_byte"),
    ENDIAN=">", SIZE=1
    )


sbsp_cluster_collision_mopp_code = Struct("collision_mopp_code",
    SInt32("unknown_0", VISIBLE=False),
    SInt16("size_0"),
    SInt16("size_1"),
    SInt32("offset_0"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    QStruct("offset_1", INCLUDE=xyz_float),
    Float("offset_scale"),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    SInt32("data_size"),
    UInt32("data_capacity"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_reflexive("data", sbsp_cluster_collision_mopp_code_data),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=64
    )


sbsp_cluster_seam = Struct("seam",
    SInt8("seam_index"),
    ENDIAN=">", SIZE=1
    )


sbsp_cluster_decorator_grid = Struct("decorator_grid",
    SInt16("amount"),
    SInt8("decorator_index"),
    SInt8("decorator_index_scattering"),
    SInt32("unknown", VISIBLE=False),
    QStruct("position", INCLUDE=xyz_float),
    Float("radius"),
    QStruct("grid_size", INCLUDE=xyz_float),
    QStruct("bounding_sphere", INCLUDE=xyz_float),
    ENDIAN=">", SIZE=48
    )


sbsp_cluster_unknown_11 = Struct("unknown_11",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_cluster_unknown_12 = Struct("unknown_12",
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    SInt16("unknown_3", VISIBLE=False),
    SInt16("unknown_4", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


sbsp_cluster = Struct("cluster",
    QStruct("bounds_x", INCLUDE=from_to),
    QStruct("bounds_y", INCLUDE=from_to),
    QStruct("bounds_z", INCLUDE=from_to),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("scenario_sky_index"),
    SInt8("camera_effect_index"),
    SInt8("unknown_1", VISIBLE=False),
    SInt16("background_sound_environment_index"),
    SInt16("sound_clusters_a_index", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=SInt16("unknown"), SIZE=5, VISIBLE=False),
    SInt16("runtime_decal_start_index"),
    SInt16("runtime_decal_entry_count"),
    SInt16("flags"),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    h3_reflexive("portals", sbsp_cluster_portal),
    SInt32("unknown_3", VISIBLE=False),
    SInt16("size_0"),
    SInt16("size_1"),
    SInt32("offset_0"),
    SInt32("unknown_4", VISIBLE=False),
    BytesRaw("unknown_5", SIZE=8, VISIBLE=False),
    h3_dependency("bsp"),
    SInt32("cluster_index"),
    SInt32("unknown_6", VISIBLE=False),
    SInt16("size_2"),
    SInt16("size_3"),
    SInt32("offset_1"),
    SInt32("unknown_7", VISIBLE=False),
    BytesRaw("unknown_8", SIZE=12, VISIBLE=False),
    h3_reflexive("collision_mopp_codes", sbsp_cluster_collision_mopp_code),
    SInt16("mesh_index"),
    SInt16("unknown_9", VISIBLE=False),
    h3_reflexive("seams", sbsp_cluster_seam),
    h3_reflexive("decorator_grids", sbsp_cluster_decorator_grid),
    BytesRaw("unknown_10", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_11", sbsp_cluster_unknown_11),
    h3_reflexive("unknown_12", sbsp_cluster_unknown_12),
    ENDIAN=">", SIZE=220
    )


sbsp_material_propertie = Struct("propertie",
    SInt16("type"),
    SInt16("int_value"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    Float("real_value"),
    ENDIAN=">", SIZE=12
    )


sbsp_material = Struct("material",
    h3_dependency("shader"),
    h3_reflexive("properties", sbsp_material_propertie),
    SInt32("unknown_0", VISIBLE=False),
    SInt8("breakable_surface_index"),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    SInt8("unknown_3", VISIBLE=False),
    ENDIAN=">", SIZE=36
    )


sbsp_sky_owner_cluster = Struct("sky_owner_cluster",
    SInt16("cluster_owner"),
    ENDIAN=">", SIZE=2
    )


sbsp_conveyor_surface = Struct("conveyor_surface",
    BytesRaw("u_i", SIZE=24, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


sbsp_breakable_surface = Struct("breakable_surface",
    BytesRaw("unknown", SIZE=32, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


sbsp_pathfinding_data_sector = Struct("sector",
    Bool16("pathfinding_sector_flags",
        "sector_walkable",
        ("sector_bsp_source", 1 << 3),
        "floor",
        ),
    SInt16("hint_index"),
    SInt32("first_link"),
    ENDIAN=">", SIZE=8
    )


sbsp_pathfinding_data_link = Struct("link",
    SInt16("vertex_1"),
    SInt16("vertex_2"),
    Bool16("link_flags", *unknown_flags_16),
    SInt16("hint_index"),
    SInt16("forward_link"),
    SInt16("reverse_link"),
    SInt16("left_sector"),
    SInt16("right_sector"),
    ENDIAN=">", SIZE=16
    )


sbsp_pathfinding_data_ref = Struct("ref",
    SInt32("node_or_sector_ref"),
    ENDIAN=">", SIZE=4
    )


sbsp_pathfinding_data_bsp_2d_node = Struct("bsp_2d_node",
    Pad(12),
    SInt32("left_child"),
    SInt32("right_child"),
    ENDIAN=">", SIZE=20
    )


sbsp_pathfinding_data_vertice = Struct("vertice",
    Pad(12),
    ENDIAN=">", SIZE=12
    )


sbsp_pathfinding_data_object_ref_unknown_1_unknown_1 = Struct("unknown_1",
    SInt32("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_pathfinding_data_object_ref_unknown_1 = Struct("unknown_1",
    Pad(8),
    h3_reflexive("unknown_1", sbsp_pathfinding_data_object_ref_unknown_1_unknown_1),
    SInt32("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


sbsp_pathfinding_data_object_ref = Struct("object_ref",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_1", sbsp_pathfinding_data_object_ref_unknown_1),
    SInt32("unknown_2", VISIBLE=False),
    SInt16("unknown_3", VISIBLE=False),
    SInt16("unknown_4", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


sbsp_pathfinding_data_pathfinding_hint = Struct("pathfinding_hint",
    SInt16("next_hint_index"),
    Pad(2),
    Array("hint_data_array", SUB_STRUCT=SInt16("hint_data"), SIZE=8),
    ENDIAN=">", SIZE=20
    )


sbsp_pathfinding_data_instanced_geometry_ref = Struct("instanced_geometry_ref",
    SInt16("pathfinding_object_index"),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=4
    )


sbsp_pathfinding_data_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_pathfinding_data_unknown_2_unknown = Struct("unknown",
    SInt32("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_pathfinding_data_unknown_2 = Struct("unknown_2",
    h3_reflexive("unknown", sbsp_pathfinding_data_unknown_2_unknown),
    VISIBLE=False,
    ENDIAN=">", SIZE=12
    )


sbsp_pathfinding_data_unknown_3_unknown_3 = Struct("unknown_3",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_pathfinding_data_unknown_3 = Struct("unknown_3",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    h3_reflexive("unknown_3", sbsp_pathfinding_data_unknown_3_unknown_3),
    VISIBLE=False,
    ENDIAN=">", SIZE=20
    )


sbsp_pathfinding_data_unknown_4 = Struct("unknown_4",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sbsp_pathfinding_data = Struct("pathfinding_data",
    h3_reflexive("sectors", sbsp_pathfinding_data_sector),
    h3_reflexive("links", sbsp_pathfinding_data_link),
    h3_reflexive("refs", sbsp_pathfinding_data_ref),
    h3_reflexive("bsp_2d_nodes", sbsp_pathfinding_data_bsp_2d_node),
    h3_reflexive("vertices", sbsp_pathfinding_data_vertice),
    h3_reflexive("object_refs", sbsp_pathfinding_data_object_ref),
    h3_reflexive("pathfinding_hints", sbsp_pathfinding_data_pathfinding_hint),
    h3_reflexive("instanced_geometry_refs", sbsp_pathfinding_data_instanced_geometry_ref),
    SInt32("structure_checksum"),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_1", sbsp_pathfinding_data_unknown_1),
    h3_reflexive("unknown_2", sbsp_pathfinding_data_unknown_2),
    h3_reflexive("unknown_3", sbsp_pathfinding_data_unknown_3),
    h3_reflexive("unknown_4", sbsp_pathfinding_data_unknown_4),
    ENDIAN=">", SIZE=160
    )


sbsp_background_sound_environment_palette = Struct("background_sound_environment_palette",
    h3_string_id("name"),
    h3_dependency("sound_environment"),
    Float("cutoff_distance_0"),
    Float("interpolation_speed_0"),
    h3_dependency("background_sound"),
    h3_dependency("inside_cluster_sound"),
    Float("cutoff_distance_1"),
    Bool32("scale_flags", *unknown_flags_32),
    Float("interior_scale"),
    Float("portal_scale"),
    Float("exterior_scale"),
    Float("interpolation_speed_1"),
    ENDIAN=">", SIZE=84
    )


sbsp_marker = Struct("marker",
    ascii_str32("name"),
    QStruct("rotation", INCLUDE=ijkw_float),
    QStruct("position", INCLUDE=xyz_float),
    ENDIAN=">", SIZE=60
    )


sbsp_light = Struct("light",
    h3_dependency("light"),
    ENDIAN=">", SIZE=16
    )


sbsp_unknown_6 = Struct("unknown_6",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


sbsp_runtime_decal = Struct("runtime_decal",
    SInt16("palette_index"),
    SInt8("yaw"),
    SInt8("pitch"),
    QStruct("rotation", INCLUDE=ijkw_float),
    QStruct("position", INCLUDE=xyz_float),
    Float("scale"),
    ENDIAN=">", SIZE=36
    )


sbsp_environment_object_palette = Struct("environment_object_palette",
    h3_dependency("definition"),
    h3_dependency("model"),
    Bool32("object_type",
        "biped",
        "vehicle",
        "weapon",
        "equipment",
        "terminal",
        "projectile",
        "scenery",
        "machine",
        "control",
        "sound_scenery",
        "crate",
        "creature",
        "giant",
        "effect_scenery",
        ),
    ENDIAN=">", SIZE=36
    )


sbsp_environment_object = Struct("environment_object",
    ascii_str32("name"),
    QStruct("rotation", INCLUDE=ijkw_float),
    QStruct("position", INCLUDE=xyz_float),
    Float("scale"),
    SInt16("palette_index"),
    SInt16("unknown_0", VISIBLE=False),
    SInt32("unique_id"),
    ascii_str32("scenario_object_name"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=108
    )


sbsp_instanced_geometry_instance_collision_definition = Struct("collision_definition",
    SInt32("unknown_0", VISIBLE=False),
    SInt16("size_0"),
    SInt16("size_1"),
    SInt32("offset_0", VISIBLE=False),
    SInt32("unknown_1", VISIBLE=False),
    BytesRaw("unknown_2", SIZE=16, VISIBLE=False),
    Float("unknown_3", VISIBLE=False),
    Float("unknown_4", VISIBLE=False),
    Float("unknown_5", VISIBLE=False),
    BytesRaw("unknown_6", SIZE=4, VISIBLE=False),
    Float("unknown_7", VISIBLE=False),
    Float("unknown_8", VISIBLE=False),
    Float("unknown_9", VISIBLE=False),
    BytesRaw("unknown_10", SIZE=4, VISIBLE=False),
    SInt32("unknown_11", VISIBLE=False),
    BytesRaw("unknown_12", SIZE=4, VISIBLE=False),
    SInt8("bsp_index"),
    SInt8("unknown_13", VISIBLE=False),
    SInt16("instanced_geometry_index"),
    Float("unknown_14", VISIBLE=False),
    SInt32("unknown_15", VISIBLE=False),
    SInt16("size_2"),
    SInt16("size_3"),
    SInt32("offset_1", VISIBLE=False),
    SInt32("unknown_16", VISIBLE=False),
    BytesRaw("unknown_17", SIZE=12, VISIBLE=False),
    Float("unknown_18", VISIBLE=False),
    ENDIAN=">", SIZE=112
    )


sbsp_instanced_geometry_instance = Struct("instanced_geometry_instance",
    Float("scale"),
    QStruct("forward", INCLUDE=ijk_float),
    QStruct("left", INCLUDE=ijk_float),
    QStruct("up", INCLUDE=ijk_float),
    QStruct("position", INCLUDE=xyz_float),
    SInt16("mesh_index"),
    Bool16("flags",
        ("no_projectile_collision", 1 << 1),
        ),
    SInt16("unknown_yo_index", VISIBLE=False),
    SInt16("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    QStruct("bounding_sphere", INCLUDE=xyz_float),
    Float("bounding_sphere_radius1"),
    Float("bounding_sphere_radius2"),
    h3_string_id("name"),
    SInt16("pathfinding_policy"),
    SInt16("lightmapping_policy"),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    h3_reflexive("collision_definitions", sbsp_instanced_geometry_instance_collision_definition),
    Array("unknown_array", SUB_STRUCT=SInt16("unknown"), SIZE=4, VISIBLE=False),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=120
    )


sbsp_decorator = Struct("decorator",
    h3_dependency("decorator"),
    ENDIAN=">", SIZE=16
    )


sbsp_meshes_0_part = Struct("part",
    SInt16("material_index"),
    SInt16("unknown_nodey_index", VISIBLE=False),
    SInt16("index_buffer_start"),
    SInt16("index_buffer_count"),
    SInt16("subpart_index"),
    SInt16("subpart_count"),
    SInt8("unknown_enum", VISIBLE=False),
    Bool8("flags",
        ("water", 1 << 3),
        ),
    SInt16("vertex_count"),
    ENDIAN=">", SIZE=16
    )


sbsp_meshes_0_subpart = Struct("subpart",
    SInt16("index_buffer_start"),
    SInt16("index_buffer_count"),
    SInt16("part_index"),
    SInt16("vertex_count"),
    ENDIAN=">", SIZE=8
    )


sbsp_meshes_0_instanced_geometry_indice_instanced_geometry_mesh_content = Struct("instanced_geometry_mesh_content",
    SInt16("instanced_geometry_index"),
    ENDIAN=">", SIZE=2
    )


sbsp_meshes_0_instanced_geometry_indice = Struct("instanced_geometry_indice",
    SInt16("instanced_geometry_mesh_index_1"),
    SInt16("instanced_geometry_mesh_index_2"),
    h3_reflexive("instanced_geometry_mesh_contents", sbsp_meshes_0_instanced_geometry_indice_instanced_geometry_mesh_content),
    ENDIAN=">", SIZE=16
    )


sbsp_meshes_0_unknown_water = Struct("unknown_water",
    SInt16("unknown"),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


sbsp_meshes_0 = Struct("meshes_0",
    h3_reflexive("parts", sbsp_meshes_0_part),
    h3_reflexive("subparts", sbsp_meshes_0_subpart),
    Array("vertex_buffer_index_array", SUB_STRUCT=SInt16("vertex_buffer_index"), SIZE=8),
    SInt16("index_buffer_index_1"),
    SInt16("index_buffer_index_2"),
    Bool8("flags",
        "has_vertex_colors",
        ),
    SInt8("rigid_node"),
    SEnum8("vertex_type", *pmdf_meshe_vertex_type),
    SEnum8("prt_type", *pmdf_meshe_prt_type),
    SEnum8("index_buffer_type", *pmdf_meshe_index_buffer_type),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    h3_reflexive("instanced_geometry_indices", sbsp_meshes_0_instanced_geometry_indice),
    h3_reflexive("unknown_water", sbsp_meshes_0_unknown_water),
    ENDIAN=">", SIZE=76
    )


sbsp_compression_info_0 = Struct("compression_info_0",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    QStruct("position_bounds_x", INCLUDE=from_to),
    QStruct("position_bounds_y", INCLUDE=from_to),
    QStruct("position_bounds_z", INCLUDE=from_to),
    QStruct("texcoord_bounds_x", INCLUDE=from_to),
    QStruct("texcoord_bounds_y", INCLUDE=from_to),
    ENDIAN=">", SIZE=44
    )


sbsp_unknown_nodey_0 = Struct("unknown_nodey_0",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    Array("node_index_array", SUB_STRUCT=SInt8("node_index"), SIZE=4, VISIBLE=False),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


sbsp_unknown_9 = Struct("unknown_9",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    h3_rawdata_ref("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


sbsp_unknown_meshes_0_unknown_1 = Struct("unknown_1",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_meshes_0 = Struct("unknown_meshes_0",
    h3_rawdata_ref("unknown_0", VISIBLE=False),
    h3_reflexive("unknown_1", sbsp_unknown_meshes_0_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


sbsp_node_maps_0_unknown = Struct("unknown",
    UInt8("node_index"),
    ENDIAN=">", SIZE=1
    )


sbsp_node_maps_0 = Struct("node_maps_0",
    h3_reflexive("unknown", sbsp_node_maps_0_unknown),
    ENDIAN=">", SIZE=12
    )


sbsp_unknown_11_unknown = Struct("unknown",
    BytesRaw("unknown", SIZE=48, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


sbsp_unknown_11 = Struct("unknown_11",
    h3_reflexive("unknown", sbsp_unknown_11_unknown),
    VISIBLE=False,
    ENDIAN=">", SIZE=12
    )


sbsp_unknown_yo_0 = Struct("unknown_yo_0",
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    SInt16("unknown_index"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


sbsp_unknown_sound_clusters_a_portal_designator = Struct("portal_designator",
    SInt16("portal_designator"),
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_sound_clusters_a_interior_cluster_indice = Struct("interior_cluster_indice",
    SInt16("interior_cluster_index"),
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_sound_clusters_a = Struct("unknown_sound_clusters_a",
    SInt16("background_sound_environment_index"),
    SInt16("unknown", VISIBLE=False),
    h3_reflexive("portal_designators", sbsp_unknown_sound_clusters_a_portal_designator),
    h3_reflexive("interior_cluster_indices", sbsp_unknown_sound_clusters_a_interior_cluster_indice),
    ENDIAN=">", SIZE=28
    )


sbsp_unknown_sound_clusters_b_portal_designator = Struct("portal_designator",
    SInt16("portal_designator"),
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_sound_clusters_b_interior_cluster_indice = Struct("interior_cluster_indice",
    SInt16("interior_cluster_index"),
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_sound_clusters_b = Struct("unknown_sound_clusters_b",
    SInt16("background_sound_environment_index"),
    SInt16("unknown", VISIBLE=False),
    h3_reflexive("portal_designators", sbsp_unknown_sound_clusters_b_portal_designator),
    h3_reflexive("interior_cluster_indices", sbsp_unknown_sound_clusters_b_interior_cluster_indice),
    ENDIAN=">", SIZE=28
    )


sbsp_unknown_sound_clusters_c_portal_designator = Struct("portal_designator",
    SInt16("portal_designator"),
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_sound_clusters_c_interior_cluster_indice = Struct("interior_cluster_indice",
    SInt16("interior_cluster_index"),
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_sound_clusters_c = Struct("unknown_sound_clusters_c",
    SInt16("background_sound_environment_index"),
    SInt16("unknown", VISIBLE=False),
    h3_reflexive("portal_designators", sbsp_unknown_sound_clusters_c_portal_designator),
    h3_reflexive("interior_cluster_indices", sbsp_unknown_sound_clusters_c_interior_cluster_indice),
    ENDIAN=">", SIZE=28
    )


sbsp_transparent_plane = Struct("transparent_plane",
    SInt16("mesh_index"),
    SInt16("part_index"),
    QStruct("plane", INCLUDE=ijk_float),
    Float("plane_d"),
    ENDIAN=">", SIZE=20
    )


sbsp_collision_mopp_code_data = Struct("data",
    UInt8("data_byte"),
    ENDIAN=">", SIZE=1
    )


sbsp_collision_mopp_code = Struct("collision_mopp_code",
    SInt32("unknown_0", VISIBLE=False),
    SInt16("size_0"),
    SInt16("size_1"),
    SInt32("offset_0"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    QStruct("offset_1", INCLUDE=xyz_float),
    Float("offset_scale"),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    SInt32("data_size"),
    UInt32("data_capacity"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_reflexive("data", sbsp_collision_mopp_code_data),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=64
    )


sbsp_breakable_surface_mopp_code_data = Struct("data",
    UInt8("data_byte"),
    ENDIAN=">", SIZE=1
    )


sbsp_breakable_surface_mopp_code = Struct("breakable_surface_mopp_code",
    SInt32("unknown_0", VISIBLE=False),
    SInt16("size_0"),
    SInt16("size_1"),
    SInt32("offset_0"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    QStruct("offset_1", INCLUDE=xyz_float),
    Float("offset_scale"),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    SInt32("data_size"),
    UInt32("data_capacity"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_reflexive("data", sbsp_breakable_surface_mopp_code_data),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=64
    )


sbsp_breakable_surface_key_table = Struct("breakable_surface_key_table",
    SInt16("instanced_geometry_index"),
    SInt8("breakable_surface_index"),
    UInt8("breakable_surface_sub_index"),
    SInt32("seed_surface_index"),
    Float("x0"),
    Float("x1"),
    Float("y0"),
    Float("y1"),
    Float("z0"),
    Float("z1"),
    ENDIAN=">", SIZE=32
    )


sbsp_meshes_1_part = Struct("part",
    SInt16("material_index"),
    SInt16("unknown_nodey_index", VISIBLE=False),
    SInt16("index_buffer_start"),
    SInt16("index_buffer_count"),
    SInt16("subpart_index"),
    SInt16("subpart_count"),
    SInt8("unknown_enum", VISIBLE=False),
    Bool8("flags",
        ("water", 1 << 3),
        ),
    SInt16("vertex_count"),
    ENDIAN=">", SIZE=16
    )


sbsp_meshes_1_subpart = Struct("subpart",
    SInt16("index_buffer_start"),
    SInt16("index_buffer_count"),
    SInt16("part_index"),
    SInt16("vertex_count"),
    ENDIAN=">", SIZE=8
    )


sbsp_meshes_1_instanced_geometry_indice_instanced_geometry_mesh_content = Struct("instanced_geometry_mesh_content",
    SInt16("instanced_geometry_index"),
    ENDIAN=">", SIZE=2
    )


sbsp_meshes_1_instanced_geometry_indice = Struct("instanced_geometry_indice",
    SInt16("instanced_geometry_mesh_index_1"),
    SInt16("instanced_geometry_mesh_index_2"),
    h3_reflexive("instanced_geometry_mesh_contents", sbsp_meshes_1_instanced_geometry_indice_instanced_geometry_mesh_content),
    ENDIAN=">", SIZE=16
    )


sbsp_meshes_1_unknown_water = Struct("unknown_water",
    SInt16("unknown"),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


sbsp_meshes_1 = Struct("meshes_1",
    h3_reflexive("parts", sbsp_meshes_1_part),
    h3_reflexive("subparts", sbsp_meshes_1_subpart),
    Array("vertex_buffer_index_array", SUB_STRUCT=SInt16("vertex_buffer_index"), SIZE=8),
    SInt16("index_buffer_index_1"),
    SInt16("index_buffer_index_2"),
    Bool8("flags",
        "has_vertex_colors",
        ),
    SInt8("rigid_node"),
    SEnum8("vertex_type", *pmdf_meshe_vertex_type),
    SEnum8("prt_type", *pmdf_meshe_prt_type),
    SEnum8("index_buffer_type", *pmdf_meshe_index_buffer_type),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    h3_reflexive("instanced_geometry_indices", sbsp_meshes_1_instanced_geometry_indice),
    h3_reflexive("unknown_water", sbsp_meshes_1_unknown_water),
    ENDIAN=">", SIZE=76
    )


sbsp_compression_info_1 = Struct("compression_info_1",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    QStruct("position_bounds_x", INCLUDE=from_to),
    QStruct("position_bounds_y", INCLUDE=from_to),
    QStruct("position_bounds_z", INCLUDE=from_to),
    QStruct("texcoord_bounds_x", INCLUDE=from_to),
    QStruct("texcoord_bounds_y", INCLUDE=from_to),
    ENDIAN=">", SIZE=44
    )


sbsp_unknown_nodey_1 = Struct("unknown_nodey_1",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    Array("node_index_array", SUB_STRUCT=SInt8("node_index"), SIZE=4, VISIBLE=False),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


sbsp_unknown_17 = Struct("unknown_17",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    h3_rawdata_ref("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


sbsp_unknown_meshes_1_unknown_1 = Struct("unknown_1",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


sbsp_unknown_meshes_1 = Struct("unknown_meshes_1",
    h3_rawdata_ref("unknown_0", VISIBLE=False),
    h3_reflexive("unknown_1", sbsp_unknown_meshes_1_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


sbsp_node_maps_1_unknown = Struct("unknown",
    UInt8("node_index"),
    ENDIAN=">", SIZE=1
    )


sbsp_node_maps_1 = Struct("node_maps_1",
    h3_reflexive("unknown", sbsp_node_maps_1_unknown),
    ENDIAN=">", SIZE=12
    )


sbsp_unknown_19_unknown = Struct("unknown",
    BytesRaw("unknown", SIZE=48, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


sbsp_unknown_19 = Struct("unknown_19",
    h3_reflexive("unknown", sbsp_unknown_19_unknown),
    VISIBLE=False,
    ENDIAN=">", SIZE=12
    )


sbsp_unknown_yo_1 = Struct("unknown_yo_1",
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    SInt16("unknown_index"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


sbsp_leaf_system = Struct("leaf_system",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    h3_dependency("leaf_system"),
    ENDIAN=">", SIZE=20
    )


sbsp_body = Struct("tagdata",
    SInt32("bsp_checksum"),
    SInt32("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    h3_reflexive("structure_seams", sbsp_structure_seam),
    h3_reflexive("unknown_raw_7th", sbsp_unknown_raw_7th),
    h3_reflexive("collision_materials", sbsp_collision_material),
    h3_reflexive("unknown_raw_3rd", sbsp_unknown_raw_3rd),
    QStruct("world_bounds_x", INCLUDE=from_to),
    QStruct("world_bounds_y", INCLUDE=from_to),
    QStruct("world_bounds_z", INCLUDE=from_to),
    h3_reflexive("unknown_raw_6th", sbsp_unknown_raw_6th),
    h3_reflexive("unknown_2", sbsp_unknown_1),
    h3_reflexive("cluster_portals", sbsp_cluster_portal),
    h3_reflexive("unknown_3", sbsp_unknown_2),
    h3_reflexive("fog", sbsp_fog),
    h3_reflexive("camera_effects", sbsp_camera_effect),
    BytesRaw("unknown_4", SIZE=12, VISIBLE=False),
    h3_reflexive("detail_objects", sbsp_detail_object),
    h3_reflexive("clusters", sbsp_cluster),
    h3_reflexive("materials", sbsp_material),
    h3_reflexive("sky_owner_cluster", sbsp_sky_owner_cluster),
    h3_reflexive("conveyor_surfaces", sbsp_conveyor_surface),
    h3_reflexive("breakable_surfaces", sbsp_breakable_surface),
    h3_reflexive("pathfinding_data", sbsp_pathfinding_data),
    BytesRaw("unknown_5", SIZE=12, VISIBLE=False),
    h3_reflexive("background_sound_environment_palette", sbsp_background_sound_environment_palette),
    BytesRaw("unknown_6", SIZE=44, VISIBLE=False),
    h3_reflexive("markers", sbsp_marker),
    h3_reflexive("lights", sbsp_light),
    h3_reflexive("unknown_7", sbsp_unknown_6),
    h3_reflexive("runtime_decals", sbsp_runtime_decal),
    h3_reflexive("environment_object_palette", sbsp_environment_object_palette),
    h3_reflexive("environment_objects", sbsp_environment_object),
    BytesRaw("unknown_8", SIZE=40, VISIBLE=False),
    h3_reflexive("instanced_geometry_instances", sbsp_instanced_geometry_instance),
    h3_reflexive("decorators", sbsp_decorator),
    SInt32("unknown_9", VISIBLE=False),
    h3_reflexive("meshes_0", sbsp_meshes_0),
    h3_reflexive("compression_info_0", sbsp_compression_info_0),
    h3_reflexive("unknown_nodey_0", sbsp_unknown_nodey_0),
    h3_reflexive("unknown_10", sbsp_unknown_9),
    BytesRaw("unknown_11", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_meshes_0", sbsp_unknown_meshes_0),
    h3_reflexive("node_maps_0", sbsp_node_maps_0),
    h3_reflexive("unknown_12", sbsp_unknown_11),
    BytesRaw("unknown_13", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_yo_0", sbsp_unknown_yo_0),
    zone_asset("asset_data_0"),
    h3_reflexive("unknown_sound_clusters_a", sbsp_unknown_sound_clusters_a),
    h3_reflexive("unknown_sound_clusters_b", sbsp_unknown_sound_clusters_b),
    h3_reflexive("unknown_sound_clusters_c", sbsp_unknown_sound_clusters_c),
    h3_reflexive("transparent_planes", sbsp_transparent_plane),
    BytesRaw("unknown_14", SIZE=12, VISIBLE=False),
    h3_reflexive("collision_mopp_codes", sbsp_collision_mopp_code),
    BytesRaw("unknown_15", SIZE=4, VISIBLE=False),
    Float("collision_world_bounds_x_min"),
    Float("collision_world_bounds_y_min"),
    Float("collision_world_bounds_z_min"),
    Float("collision_world_bounds_x_max"),
    Float("collision_world_bounds_y_max"),
    Float("collision_world_bounds_z_max"),
    h3_reflexive("breakable_surface_mopp_codes", sbsp_breakable_surface_mopp_code),
    h3_reflexive("breakable_surface_key_table", sbsp_breakable_surface_key_table),
    BytesRaw("unknown_16", SIZE=24, VISIBLE=False),
    SInt32("unknown_17", VISIBLE=False),
    h3_reflexive("meshes_1", sbsp_meshes_1),
    h3_reflexive("compression_info_1", sbsp_compression_info_1),
    h3_reflexive("unknown_nodey_1", sbsp_unknown_nodey_1),
    h3_reflexive("unknown_18", sbsp_unknown_17),
    BytesRaw("unknown_19", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_meshes_1", sbsp_unknown_meshes_1),
    h3_reflexive("node_maps_1", sbsp_node_maps_1),
    h3_reflexive("unknown_20", sbsp_unknown_19),
    BytesRaw("unknown_21", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_yo_1", sbsp_unknown_yo_1),
    zone_asset("asset_data_1"),
    h3_reflexive("leaf_systems", sbsp_leaf_system),
    BytesRaw("unknown_22", SIZE=12, VISIBLE=False),
    UInt16("zone_asset_salt_3"),
    UInt16("zone_asset_index_3"),
    SInt32("useless_padding_2", VISIBLE=False),
    SInt32("unknown_23", VISIBLE=False),
    ENDIAN=">", SIZE=904
    )


def get():
    return sbsp_def

sbsp_def = TagDef("sbsp",
    h3_blam_header('sbsp'),
    sbsp_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["sbsp"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.h2.common_descs import *
from supyr_struct.defs.tag_def import TagDef

vertex = Struct("vertex",
    Float("spring strength coefficient"),

    Pad(24),
    yp_float_rad("angles"),  # radians
    float_wu("length"),
    SInt16("sequence index"),

    Pad(2),
    QStruct("color", INCLUDE=argb_float),
    QStruct("lod color", INCLUDE=argb_float),
    SIZE=128
    )

ant__body = Struct("tagdata",
    h2_string_id("attachment marker name"),
    h2_dependency("bitmaps", "bitm"),
    h2_dependency("physics", "pphy"),

    Pad(80),
    Float("spring strength coefficient"),
    Float("falloff pixels"),
    Float("cutoff pixels"),

    Pad(40),
    h2_reflexive("vertices", vertex, 20),
    ENDIAN="<", SIZE=160
    )


def get():
    return ant__def

ant__def = TagDef("ant!",
    h2_blam_header('ant!'),
    ant__body,

    ext=".%s" % h2_tag_class_fcc_to_ext["ant!"], endian="<"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get():
    return unit_def

camera_track = Struct('camera_track',
    dependency('track', "trak"),
    SIZE=28
    )

new_hud_interface = Struct('new_hud_interface',
    dependency('unit_hud_interface', "unhi"),
    SIZE=48
    )

dialogue_variant = Struct('dialogue_variant',
    SInt16('variant_number'),
    Pad(6),
    dependency('dialogue', "udlg"),
    SIZE=24
    )

powered_seat = Struct('powered_seat',
    Pad(4),
    float_sec('driver_powerup_time'),
    float_sec('driver_powerdown_time'),
    SIZE=68
    )

weapon = Struct('weapon',
    dependency('weapon', "weap"),
    SIZE=36
    )

seat = Struct('seat',
    Bool32("flags",
        "invisible",
        "locked",
        "driver",
        "gunner",
        "third_person_camera",
        "allows_weapons",
        "third_person_on_enter",
        "first_person_slaved_to_gun",
        "allow_vehicle_communcation_animation",
        "not_valid_without_driver",
        "allow_ai_noncombatants"
        ),
    ascii_str32('label'),
    ascii_str32('marker_name'),

    Pad(32),
    QStruct("acceleration_scale", INCLUDE=ijk_float),

    Pad(12),
    float_deg_sec('yaw_rate',   UNIT_SCALE=per_sec_unit_scale),  # degrees/sec
    float_deg_sec('pitch_rate', UNIT_SCALE=per_sec_unit_scale),  # degrees/sec
    ascii_str32('camera_marker_name'),
    ascii_str32('camera_submerged_marker_name'),
    float_rad('pitch_auto_level'),  # radians
    from_to_rad('pitch_range'),  # radians

    reflexive("camera_tracks", camera_track, 2, 'loose', 'tight'),
    reflexive("new_hud_interfaces", new_hud_interface, 2,
              'default/solo', 'multiplayer'),

    Pad(4),
    SInt16("hud_text_message_index"),

    Pad(2),
    float_rad('yaw_minimum'),  # radians
    float_rad('yaw_maximum'),  # radians
    dependency('built_in_gunner', "actv"),
    Pad(12),  # open sauce seat extension padding
    SIZE=284
    )

unit_attrs = Struct("unit_attrs",
    Bool32("flags",
        "circular_aiming",
        "destroyed_after_dying",
        "half_speed_interpolation",
        "fires_from_camera",
        "entrance_inside_bounding_sphere",
        "unused",
        "causes_passenger_dialogue",
        "resists_pings",
        "melee_attack_is_fatal",
        "dont_reface_during_pings",
        "has_no_aiming",
        "simple_creature",
        "impact_melee_attaches_to_unit",
        "impact_melee_dies_on_shields",
        "cannot_open_doors_automatically",
        "melee_attackers_cannot_attach",
        "not_instantly_killed_by_melee",
        "shield_sapping",
        "runs_around_flaming",
        "inconsequential",
        "special_cinematic_unit",
        "ignored_by_autoaiming",
        "shields_fry_infection_forms",
        "integrated_light_controls_weapon",
        "integrated_light_lasts_forever",
        ),
    SEnum16('default_team', *unit_teams),
    SEnum16('constant_sound_volume', *sound_volumes),
    float_zero_to_inf('rider_damage_fraction'),
    dependency('integrated_light_toggle', "effe"),
    SEnum16('A_in', *unit_inputs),
    SEnum16('B_in', *unit_inputs),
    SEnum16('C_in', *unit_inputs),
    SEnum16('D_in', *unit_inputs),
    float_rad('camera_field_of_view'),  # radians
    Float('camera_stiffness'),
    ascii_str32('camera_marker_name'),
    ascii_str32('camera_submerged_marker_name'),
    float_rad('pitch_auto_level'),  # radians
    from_to_rad('pitch_range'),  # radians
    reflexive("camera_tracks", camera_track, 2,
              'loose', 'tight'),

    #Miscellaneous
    QStruct("seat_acceleration_scale", INCLUDE=ijk_float),
    Pad(12),
    float_zero_to_one('soft_ping_threshold'),  # [0,1]
    float_sec('soft_ping_interrupt_time', UNIT_SCALE=sec_unit_scale),  # seconds
    float_zero_to_one('hard_ping_threshold'),  # [0,1]
    float_sec('hard_ping_interrupt_time', UNIT_SCALE=sec_unit_scale),  # seconds
    float_zero_to_one('hard_death_threshold'),  # [0,1]
    float_zero_to_one('feign_death_threshold'),  # [0,1]
    float_sec('feign_death_time', UNIT_SCALE=sec_unit_scale),  # seconds
    float_wu('distance_of_evade_aim'),  # world units
    float_wu('distance_of_dive_aim'),  # world units

    Pad(4),
    float_zero_to_one('stunned_movement_threshold'),  # [0,1]
    float_zero_to_one('feign_death_chance'),  # [0,1]
    float_zero_to_one('feign_repeat_chance'),  # [0,1]
    dependency('spawned_actor', "actv"),
    QStruct("spawned_actor_count",
        SInt16("from", GUI_NAME=""), SInt16("to"), ORIENT='h',
        ),
    float_wu_sec('spawned_velocity'),
    float_rad_sec('aiming_velocity_maximum',
                  UNIT_SCALE=irad_per_sec_unit_scale),  # radians/sec
    float_rad_sec_sq('aiming_acceleration_maximum',
                     UNIT_SCALE=irad_per_sec_sq_unit_scale),  # radians/sec^2
    float_zero_to_one('casual_aiming_modifier'),
    float_rad_sec('looking_velocity_maximum',
                  UNIT_SCALE=irad_per_sec_unit_scale),  # radians/sec
    float_rad_sec_sq('looking_acceleration_maximum',
                     UNIT_SCALE=irad_per_sec_sq_unit_scale),  # radians/sec^2

    Pad(8),
    Float('ai_vehicle_radius'),
    Float('ai_danger_radius'),
    dependency('melee_damage', "jpt!"),
    SEnum16('motion_sensor_blip_size',
        "medium",
        "small",
        "large",
        ),
    Pad(2),

    Pad(12),  # open sauce unit extension padding
    reflexive("new_hud_interfaces", new_hud_interface, 2,
        'default/solo', 'multiplayer'),
    reflexive("dialogue_variants", dialogue_variant, 16,
        DYN_NAME_PATH='.dialogue.filepath'),

    #Grenades
    float_wu_sec('grenade_velocity'),
    SEnum16('grenade_type', *grenade_types),
    SInt16('grenade_count', MIN=0),

    Pad(4),
    reflexive("powered_seats", powered_seat, 2,
              "driver", "gunner"),
    reflexive("weapons", weapon, 4, DYN_NAME_PATH='.weapon.filepath'),
    reflexive("seats", seat, 16, DYN_NAME_PATH='.label'),

    SIZE=372
    )

unit_body = Struct('tagdata',
    unit_attrs,
    SIZE=372
    )

unit_def = TagDef("unit",
    blam_header('unit', 2),
    unit_body,

    ext=".unit", endian=">", tag_cls=HekTag
    )
